import json
from typing import Any
from urllib.parse import unquote

import xmltodict

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import (
    ActionResult,
    BaseResponse,
    EmptyResult,
)
from moto.s3.responses import S3_PUBLIC_ACCESS_BLOCK_CONFIGURATION

from .models import S3ControlBackend, s3control_backends


class S3ControlResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="s3control")

    @property
    def backend(self) -> S3ControlBackend:
        return s3control_backends[self.current_account][self.partition]

    def get_public_access_block(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        public_block_config = self.backend.get_public_access_block(
            account_id=account_id
        )
        template = self.response_template(S3_PUBLIC_ACCESS_BLOCK_CONFIGURATION)
        return template.render(public_block_config=public_block_config)

    def put_public_access_block(self) -> TYPE_RESPONSE:
        account_id = self.headers.get("x-amz-account-id")
        pab_config = self._parse_pab_config(self.body)
        self.backend.put_public_access_block(
            account_id, pab_config["PublicAccessBlockConfiguration"]
        )
        return 201, {"status": 201}, json.dumps({})

    def delete_public_access_block(self) -> TYPE_RESPONSE:
        account_id = self.headers.get("x-amz-account-id")
        self.backend.delete_public_access_block(account_id=account_id)
        return 204, {"status": 204}, json.dumps({})

    def _parse_pab_config(self, body: str) -> dict[str, Any]:
        parsed_xml = xmltodict.parse(body)
        parsed_xml["PublicAccessBlockConfiguration"].pop("@xmlns", None)

        return parsed_xml

    def create_access_point(self) -> str:
        account_id, name = self._get_accountid_and_name_from_accesspoint(self.uri)
        params = xmltodict.parse(self.body)["CreateAccessPointRequest"]
        bucket = params["Bucket"]
        vpc_configuration = params.get("VpcConfiguration")
        public_access_block_configuration = params.get("PublicAccessBlockConfiguration")
        access_point = self.backend.create_access_point(
            account_id=account_id,
            name=name,
            bucket=bucket,
            vpc_configuration=vpc_configuration,
            public_access_block_configuration=public_access_block_configuration,
        )
        template = self.response_template(CREATE_ACCESS_POINT_TEMPLATE)
        return template.render(access_point=access_point)

    def get_access_point(self) -> str:
        account_id, name = self._get_accountid_and_name_from_accesspoint(self.uri)

        access_point = self.backend.get_access_point(account_id=account_id, name=name)
        template = self.response_template(GET_ACCESS_POINT_TEMPLATE)
        return template.render(access_point=access_point)

    def delete_access_point(self) -> TYPE_RESPONSE:
        account_id, name = self._get_accountid_and_name_from_accesspoint(self.uri)
        self.backend.delete_access_point(account_id=account_id, name=name)
        return 204, {"status": 204}, ""

    def put_access_point_policy(self) -> str:
        account_id, name = self._get_accountid_and_name_from_policy(self.uri)
        params = xmltodict.parse(self.body)
        policy = params["PutAccessPointPolicyRequest"]["Policy"]
        self.backend.put_access_point_policy(account_id, name, policy)
        return ""

    def get_access_point_policy(self) -> str:
        account_id, name = self._get_accountid_and_name_from_policy(self.uri)
        policy = self.backend.get_access_point_policy(account_id, name)
        template = self.response_template(GET_ACCESS_POINT_POLICY_TEMPLATE)
        return template.render(policy=policy)

    def delete_access_point_policy(self) -> TYPE_RESPONSE:
        account_id, name = self._get_accountid_and_name_from_policy(self.uri)
        self.backend.delete_access_point_policy(account_id=account_id, name=name)
        return 204, {"status": 204}, ""

    def get_access_point_policy_status(self) -> str:
        account_id, name = self._get_accountid_and_name_from_policy(self.uri)
        self.backend.get_access_point_policy_status(account_id, name)
        template = self.response_template(GET_ACCESS_POINT_POLICY_STATUS_TEMPLATE)
        return template.render()

    def _get_accountid_and_name_from_accesspoint(
        self, full_url: str
    ) -> tuple[str, str]:
        url = full_url
        if full_url.startswith("http"):
            url = full_url.split("://")[1]
        account_id = url.split(".")[0]
        name = url.split("v20180820/accesspoint/")[-1]
        return account_id, name

    def _get_accountid_and_name_from_policy(self, full_url: str) -> tuple[str, str]:
        url = full_url
        if full_url.startswith("http"):
            url = full_url.split("://")[1]
        account_id = url.split(".")[0]
        name = self.path.split("/")[-2]
        return account_id, name

    def put_storage_lens_configuration(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        config_id = self.path.split("/")[-1]
        request = xmltodict.parse(self.body)["PutStorageLensConfigurationRequest"]
        storage_lens_configuration = request.get("StorageLensConfiguration")
        tags = request.get("Tags")
        self.backend.put_storage_lens_configuration(
            config_id=config_id,
            account_id=account_id,
            storage_lens_configuration=storage_lens_configuration,
            tags=tags,
        )
        return ""

    def get_storage_lens_configuration(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        config_id = self.path.split("/")[-1]
        storage_lens_configuration = self.backend.get_storage_lens_configuration(
            config_id=config_id,
            account_id=account_id,
        )
        template = self.response_template(GET_STORAGE_LENS_CONFIGURATION_TEMPLATE)
        return template.render(config=storage_lens_configuration.config)

    def delete_storage_lens_configuration(self) -> TYPE_RESPONSE:
        account_id = self.headers.get("x-amz-account-id")
        config_id = self.path.split("/")[-1]
        self.backend.delete_storage_lens_configuration(
            config_id=config_id,
            account_id=account_id,
        )
        return 204, {"status": 204}, ""

    def list_storage_lens_configurations(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        params = self._get_params()
        next_token = params.get("nextToken")
        storage_lens_configuration_list, next_token = (
            self.backend.list_storage_lens_configurations(
                account_id=account_id,
                next_token=next_token,
            )
        )
        template = self.response_template(LIST_STORAGE_LENS_CONFIGURATIONS_TEMPLATE)
        return template.render(
            next_token=next_token, configs=storage_lens_configuration_list
        )

    def put_storage_lens_configuration_tagging(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        config_id = self.path.split("/")[-2]
        request = xmltodict.parse(self.body)[
            "PutStorageLensConfigurationTaggingRequest"
        ]
        tags = request.get("Tags")
        self.backend.put_storage_lens_configuration_tagging(
            config_id=config_id,
            account_id=account_id,
            tags=tags,
        )
        return ""

    def get_storage_lens_configuration_tagging(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        config_id = self.path.split("/")[-2]
        storage_lens_tags = self.backend.get_storage_lens_configuration_tagging(
            config_id=config_id,
            account_id=account_id,
        )
        template = self.response_template(
            GET_STORAGE_LENS_CONFIGURATION_TAGGING_TEMPLATE
        )
        return template.render(tags=storage_lens_tags)

    def list_access_points(self) -> str:
        account_id = self.headers.get("x-amz-account-id")

        params = self._get_params()
        max_results = params.get("maxResults")
        if max_results:
            max_results = int(max_results)

        access_points, next_token = self.backend.list_access_points(
            account_id=account_id,
            bucket=params.get("bucket"),
            max_results=max_results,
            next_token=params.get("nextToken"),
        )

        template = self.response_template(LIST_ACCESS_POINTS_TEMPLATE)
        return template.render(access_points=access_points, next_token=next_token)

    def create_multi_region_access_point(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        params = xmltodict.parse(self.body, process_namespaces=False)[
            "CreateMultiRegionAccessPointRequest"
        ]

        details = params["Details"]
        name = details.get("Name")

        regions_data = details.get("Regions", {})
        if regions_data and "Region" in regions_data:
            regions_list = regions_data["Region"]
            if isinstance(regions_list, dict):
                regions_list = [regions_list]
        else:
            regions_list = []

        public_access_block = details.get("PublicAccessBlock", {})

        operation = self.backend.create_multi_region_access_point(
            account_id=account_id,
            name=name,
            public_access_block=public_access_block,
            regions=regions_list,
            region_name=self.region,
        )

        template = self.response_template(CREATE_MULTI_REGION_ACCESS_POINT_TEMPLATE)
        return template.render(request_token=operation.request_token_arn)

    def delete_multi_region_access_point(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        params = xmltodict.parse(self.body, process_namespaces=False)[
            "DeleteMultiRegionAccessPointRequest"
        ]

        details = params["Details"]
        name = details.get("Name")

        operation = self.backend.delete_multi_region_access_point(
            account_id=account_id,
            name=name,
            region_name=self.region,
        )

        template = self.response_template(DELETE_MULTI_REGION_ACCESS_POINT_TEMPLATE)
        return template.render(request_token=operation.request_token_arn)

    def describe_multi_region_access_point_operation(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        _prefix = "/async-requests/mrap/"
        if _prefix in self.path:
            request_token = unquote(self.path.partition(_prefix)[2])
        else:
            request_token = self.path.split("/")[-1]

        operation = self.backend.describe_multi_region_access_point_operation(
            account_id=account_id,
            request_token_arn=request_token,
        )

        template = self.response_template(
            DESCRIBE_MULTI_REGION_ACCESS_POINT_OPERATION_TEMPLATE
        )
        return template.render(operation=operation.to_dict())

    def get_multi_region_access_point(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        name = self.path.split("/")[-1]

        mrap = self.backend.get_multi_region_access_point(
            account_id=account_id,
            name=name,
        )

        template = self.response_template(GET_MULTI_REGION_ACCESS_POINT_TEMPLATE)
        return template.render(mrap=mrap.to_dict())

    def get_multi_region_access_point_policy(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        name = self.path.split("/")[-2]

        policy = self.backend.get_multi_region_access_point_policy(
            account_id=account_id,
            name=name,
        )

        template = self.response_template(GET_MULTI_REGION_ACCESS_POINT_POLICY_TEMPLATE)
        return template.render(policy=policy)

    def get_multi_region_access_point_policy_status(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        name = self.path.split("/")[-2]

        policy_status = self.backend.get_multi_region_access_point_policy_status(
            account_id=account_id,
            name=name,
        )

        template = self.response_template(
            GET_MULTI_REGION_ACCESS_POINT_POLICY_STATUS_TEMPLATE
        )
        return template.render(is_public=policy_status["IsPublic"])

    def list_multi_region_access_points(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        params = self._get_params()

        max_results = params.get("maxResults")
        if max_results:
            max_results = int(max_results)

        mraps, next_token = self.backend.list_multi_region_access_points(
            account_id=account_id,
            max_results=max_results,
            next_token=params.get("nextToken"),
        )

        template = self.response_template(LIST_MULTI_REGION_ACCESS_POINTS_TEMPLATE)
        return template.render(
            mraps=[mrap.to_dict() for mrap in mraps], next_token=next_token
        )

    def put_multi_region_access_point_policy(self) -> str:
        account_id = self.headers.get("x-amz-account-id")
        params = xmltodict.parse(self.body, process_namespaces=False)[
            "PutMultiRegionAccessPointPolicyRequest"
        ]

        details = params["Details"]
        name = details.get("Name")
        policy = details.get("Policy")

        operation = self.backend.put_multi_region_access_point_policy(
            account_id=account_id,
            name=name,
            policy=policy,
            region_name=self.region,
        )

        template = self.response_template(PUT_MULTI_REGION_ACCESS_POINT_POLICY_TEMPLATE)
        return template.render(request_token=operation.request_token_arn)

    def list_tags_for_resource(self) -> ActionResult:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = self.backend.list_tags_for_resource(resource_arn)
        return ActionResult(result={"Tags": tags})

    def tag_resource(self) -> EmptyResult:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = (
            xmltodict.parse(self.raw_body, force_list={"Tag": True})
            .get("TagResourceRequest", {})
            .get("Tags", {})["Tag"]
        )
        self.backend.tag_resource(resource_arn, tags=tags)
        return EmptyResult()

    def untag_resource(self) -> EmptyResult:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tag_keys = self.querystring.get("tagKeys", [])
        self.backend.untag_resource(resource_arn, tag_keys=tag_keys)
        return EmptyResult()


CREATE_ACCESS_POINT_TEMPLATE = """<CreateAccessPointResult>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <Alias>{{ access_point.alias }}</Alias>
  <AccessPointArn>{{ access_point.arn }}</AccessPointArn>
</CreateAccessPointResult>
"""


GET_ACCESS_POINT_TEMPLATE = """<GetAccessPointResult>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <Name>{{ access_point.name }}</Name>
  <Bucket>{{ access_point.bucket }}</Bucket>
  <NetworkOrigin>{{ access_point.network_origin }}</NetworkOrigin>
  {% if access_point.vpc_id %}
  <VpcConfiguration>
      <VpcId>{{ access_point.vpc_id }}</VpcId>
  </VpcConfiguration>
  {% endif %}
  <PublicAccessBlockConfiguration>
      <BlockPublicAcls>{{ access_point.pubc["BlockPublicAcls"] }}</BlockPublicAcls>
      <IgnorePublicAcls>{{ access_point.pubc["IgnorePublicAcls"] }}</IgnorePublicAcls>
      <BlockPublicPolicy>{{ access_point.pubc["BlockPublicPolicy"] }}</BlockPublicPolicy>
      <RestrictPublicBuckets>{{ access_point.pubc["RestrictPublicBuckets"] }}</RestrictPublicBuckets>
  </PublicAccessBlockConfiguration>
  <CreationDate>{{ access_point.created }}</CreationDate>
  <Alias>{{ access_point.alias }}</Alias>
  <AccessPointArn>{{ access_point.arn }}</AccessPointArn>
  <Endpoints>
      <entry>
          <key>ipv4</key>
          <value>s3-accesspoint.us-east-1.amazonaws.com</value>
      </entry>
      <entry>
          <key>fips</key>
          <value>s3-accesspoint-fips.us-east-1.amazonaws.com</value>
      </entry>
      <entry>
          <key>fips_dualstack</key>
          <value>s3-accesspoint-fips.dualstack.us-east-1.amazonaws.com</value>
      </entry>
      <entry>
          <key>dualstack</key>
          <value>s3-accesspoint.dualstack.us-east-1.amazonaws.com</value>
      </entry>
  </Endpoints>
</GetAccessPointResult>
"""


GET_ACCESS_POINT_POLICY_TEMPLATE = """<GetAccessPointPolicyResult>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <Policy>{{ policy }}</Policy>
</GetAccessPointPolicyResult>
"""


GET_ACCESS_POINT_POLICY_STATUS_TEMPLATE = """<GetAccessPointPolicyResult>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <PolicyStatus>
      <IsPublic>true</IsPublic>
  </PolicyStatus>
</GetAccessPointPolicyResult>
"""

XMLNS = 'xmlns="http://awss3control.amazonaws.com/doc/2018-08-20/"'

CREATE_MULTI_REGION_ACCESS_POINT_TEMPLATE = f"""<CreateMultiRegionAccessPointResult {XMLNS}>
  <RequestTokenARN>{{{{ request_token }}}}</RequestTokenARN>
</CreateMultiRegionAccessPointResult>
"""

DELETE_MULTI_REGION_ACCESS_POINT_TEMPLATE = f"""<DeleteMultiRegionAccessPointResult {XMLNS}>
  <RequestTokenARN>{{{{ request_token }}}}</RequestTokenARN>
</DeleteMultiRegionAccessPointResult>
"""

DESCRIBE_MULTI_REGION_ACCESS_POINT_OPERATION_TEMPLATE = f"""<DescribeMultiRegionAccessPointOperationResult {XMLNS}>
  <AsyncOperation>
    <CreationTime>{{{{ operation.CreationTime }}}}</CreationTime>
    <Operation>{{{{ operation.Operation }}}}</Operation>
    <RequestTokenARN>{{{{ operation.RequestTokenARN }}}}</RequestTokenARN>
    <RequestStatus>{{{{ operation.RequestStatus }}}}</RequestStatus>
    {{% if operation.RequestParameters %}}
    <RequestParameters>
      {{% if operation.RequestParameters.CreateMultiRegionAccessPointRequest %}}
      <CreateMultiRegionAccessPointRequest>
        <Name>{{{{ operation.RequestParameters.CreateMultiRegionAccessPointRequest.Name }}}}</Name>
        {{% if operation.RequestParameters.CreateMultiRegionAccessPointRequest.PublicAccessBlock %}}
        <PublicAccessBlock>
          <BlockPublicAcls>{{{{ operation.RequestParameters.CreateMultiRegionAccessPointRequest.PublicAccessBlock.BlockPublicAcls|default("true") }}}}</BlockPublicAcls>
          <IgnorePublicAcls>{{{{ operation.RequestParameters.CreateMultiRegionAccessPointRequest.PublicAccessBlock.IgnorePublicAcls|default("true") }}}}</IgnorePublicAcls>
          <BlockPublicPolicy>{{{{ operation.RequestParameters.CreateMultiRegionAccessPointRequest.PublicAccessBlock.BlockPublicPolicy|default("true") }}}}</BlockPublicPolicy>
          <RestrictPublicBuckets>{{{{ operation.RequestParameters.CreateMultiRegionAccessPointRequest.PublicAccessBlock.RestrictPublicBuckets|default("true") }}}}</RestrictPublicBuckets>
        </PublicAccessBlock>
        {{% endif %}}
        {{% if operation.RequestParameters.CreateMultiRegionAccessPointRequest.Regions %}}
        <Regions>
          {{% for region in operation.RequestParameters.CreateMultiRegionAccessPointRequest.Regions %}}
          <Region>
            <Bucket>{{{{ region.Bucket }}}}</Bucket>
          </Region>
          {{% endfor %}}
        </Regions>
        {{% endif %}}
      </CreateMultiRegionAccessPointRequest>
      {{% endif %}}
      {{% if operation.RequestParameters.DeleteMultiRegionAccessPointRequest %}}
      <DeleteMultiRegionAccessPointRequest>
        <Name>{{{{ operation.RequestParameters.DeleteMultiRegionAccessPointRequest.Name }}}}</Name>
      </DeleteMultiRegionAccessPointRequest>
      {{% endif %}}
      {{% if operation.RequestParameters.PutMultiRegionAccessPointPolicyRequest %}}
      <PutMultiRegionAccessPointPolicyRequest>
        <Name>{{{{ operation.RequestParameters.PutMultiRegionAccessPointPolicyRequest.Name }}}}</Name>
        <Policy>{{{{ operation.RequestParameters.PutMultiRegionAccessPointPolicyRequest.Policy }}}}</Policy>
      </PutMultiRegionAccessPointPolicyRequest>
      {{% endif %}}
    </RequestParameters>
    {{% endif %}}
    {{% if operation.ResponseDetails %}}
    <ResponseDetails>
      {{% if operation.ResponseDetails.MultiRegionAccessPointDetails %}}
      <MultiRegionAccessPointDetails>
        {{% if operation.ResponseDetails.MultiRegionAccessPointDetails.Regions %}}
        <Regions>
          {{% for region in operation.ResponseDetails.MultiRegionAccessPointDetails.Regions %}}
          <Region>
            <Name>{{{{ region.Name }}}}</Name>
            <RequestStatus>{{{{ region.RequestStatus }}}}</RequestStatus>
          </Region>
          {{% endfor %}}
        </Regions>
        {{% endif %}}
      </MultiRegionAccessPointDetails>
      {{% endif %}}
    </ResponseDetails>
    {{% endif %}}
  </AsyncOperation>
</DescribeMultiRegionAccessPointOperationResult>
"""

GET_MULTI_REGION_ACCESS_POINT_TEMPLATE = f"""<GetMultiRegionAccessPointResult {XMLNS}>
  <AccessPoint>
    <Name>{{{{ mrap.Name }}}}</Name>
    <Alias>{{{{ mrap.Alias }}}}</Alias>
    <CreatedAt>{{{{ mrap.CreatedAt }}}}</CreatedAt>
    <PublicAccessBlock>
      <BlockPublicAcls>{{{{ mrap.PublicAccessBlock.BlockPublicAcls|default("true") }}}}</BlockPublicAcls>
      <IgnorePublicAcls>{{{{ mrap.PublicAccessBlock.IgnorePublicAcls|default("true") }}}}</IgnorePublicAcls>
      <BlockPublicPolicy>{{{{ mrap.PublicAccessBlock.BlockPublicPolicy|default("true") }}}}</BlockPublicPolicy>
      <RestrictPublicBuckets>{{{{ mrap.PublicAccessBlock.RestrictPublicBuckets|default("true") }}}}</RestrictPublicBuckets>
    </PublicAccessBlock>
    <Status>{{{{ mrap.Status }}}}</Status>
    <Regions>
      {{% for region in mrap.Regions %}}
      <Region>
        <Bucket>{{{{ region.Bucket }}}}</Bucket>
        <Region>{{{{ region.Region }}}}</Region>
      </Region>
      {{% endfor %}}
    </Regions>
  </AccessPoint>
</GetMultiRegionAccessPointResult>
"""

GET_MULTI_REGION_ACCESS_POINT_POLICY_TEMPLATE = f"""<GetMultiRegionAccessPointPolicyResult {XMLNS}>
  <Policy>
    <Established>
      <Policy>{{{{ policy }}}}</Policy>
    </Established>
  </Policy>
</GetMultiRegionAccessPointPolicyResult>
"""

GET_MULTI_REGION_ACCESS_POINT_POLICY_STATUS_TEMPLATE = f"""<GetMultiRegionAccessPointPolicyStatusResult {XMLNS}>
  <Established>
    <IsPublic>{{{{ is_public|lower }}}}</IsPublic>
  </Established>
</GetMultiRegionAccessPointPolicyStatusResult>
"""

LIST_MULTI_REGION_ACCESS_POINTS_TEMPLATE = f"""<ListMultiRegionAccessPointsResult {XMLNS}>
  <AccessPoints>
    {{% for mrap in mraps %}}
    <AccessPoint>
      <Name>{{{{ mrap.Name }}}}</Name>
      <Alias>{{{{ mrap.Alias }}}}</Alias>
      <CreatedAt>{{{{ mrap.CreatedAt }}}}</CreatedAt>
      <PublicAccessBlock>
        <BlockPublicAcls>{{{{ mrap.PublicAccessBlock.BlockPublicAcls|default("true") }}}}</BlockPublicAcls>
        <IgnorePublicAcls>{{{{ mrap.PublicAccessBlock.IgnorePublicAcls|default("true") }}}}</IgnorePublicAcls>
        <BlockPublicPolicy>{{{{ mrap.PublicAccessBlock.BlockPublicPolicy|default("true") }}}}</BlockPublicPolicy>
        <RestrictPublicBuckets>{{{{ mrap.PublicAccessBlock.RestrictPublicBuckets|default("true") }}}}</RestrictPublicBuckets>
      </PublicAccessBlock>
      <Status>{{{{ mrap.Status }}}}</Status>
      <Regions>
        {{% for region in mrap.Regions %}}
        <Region>
          <Bucket>{{{{ region.Bucket }}}}</Bucket>
          <Region>{{{{ region.Region }}}}</Region>
        </Region>
        {{% endfor %}}
      </Regions>
    </AccessPoint>
    {{% endfor %}}
  </AccessPoints>
  {{% if next_token %}}
  <NextToken>{{{{ next_token }}}}</NextToken>
  {{% endif %}}
</ListMultiRegionAccessPointsResult>
"""

PUT_MULTI_REGION_ACCESS_POINT_POLICY_TEMPLATE = f"""<PutMultiRegionAccessPointPolicyResult {XMLNS}>
  <RequestTokenARN>{{{{ request_token }}}}</RequestTokenARN>
</PutMultiRegionAccessPointPolicyResult>
"""

GET_STORAGE_LENS_CONFIGURATION_TEMPLATE = """
<StorageLensConfiguration>
   <Id>{{config.get("Id")}}</Id>
   {% if config.get("DataExport") %}
   <DataExport>
      {% if config["DataExport"]["S3BucketDestination"] %}
      <S3BucketDestination>
         <AccountId>{{config["DataExport"]["S3BucketDestination"]["AccountId"]}}</AccountId>
         <Arn>{{config["DataExport"]["S3BucketDestination"]["Arn"]}}</Arn>
         {% if config["DataExport"]["S3BucketDestination"].get("Encryption") %}
         <Encryption>
            {% if config["DataExport"]["S3BucketDestination"]["Encryption"].get("SSEKMS") %}
            <SSE-KMS>
               <KeyId>config["DataExport"]["S3BucketDestination"]["Encryption"]["KeyId"]</KeyId>
            </SSE-KMS>
            {% endif %}
            {% if "SSE-S3" in config["DataExport"]["S3BucketDestination"]["Encryption"] %}
            <SSE-S3>
            </SSE-S3>
            {% endif %}
         </Encryption>
         {% endif %}
      </S3BucketDestination>
      {% endif %}
   </DataExport>
   {% endif %}
   <IsEnabled>{{config["IsEnabled"]}}</IsEnabled>
   <AccountLevel>
        <ActivityMetrics>
            <IsEnabled>{{config["AccountLevel"]["ActivityMetrics"]["IsEnabled"]}}</IsEnabled>
        </ActivityMetrics>
        <BucketLevel>
            <ActivityMetrics>
                <IsEnabled>{{config["AccountLevel"]["BucketLevel"]["ActivityMetrics"]["IsEnabled"]}}</IsEnabled>
            </ActivityMetrics>
            <PrefixLevel>
                <StorageMetrics>
                    <IsEnabled>{{config["AccountLevel"]["BucketLevel"]["PrefixLevel"]["StorageMetrics"]["IsEnabled"]}}</IsEnabled>
                    <SelectionCriteria>
                        <Delimiter>{{config["AccountLevel"]["BucketLevel"]["PrefixLevel"]["StorageMetrics"]["SelectionCriteria"]["Delimiter"]}}</Delimiter>
                        <MaxDepth>{{config["AccountLevel"]["BucketLevel"]["PrefixLevel"]["StorageMetrics"]["SelectionCriteria"]["MaxDepth"]}}</MaxDepth>
                        <MinStorageBytesPercentage>{{config["AccountLevel"]["BucketLevel"]["PrefixLevel"]["StorageMetrics"]["SelectionCriteria"]["MinStorageBytesPercentage"]}}</MinStorageBytesPercentage>
                    </SelectionCriteria>
                </StorageMetrics>
            </PrefixLevel>
            <DetailedStatusCodesMetrics>
                <IsEnabled>{{config["AccountLevel"]["BucketLevel"]["DetailedStatusCodesMetrics"]["IsEnabled"]}}</IsEnabled>
            </DetailedStatusCodesMetrics>
        </BucketLevel>
        <AdvancedDataProtectionMetrics>
            <IsEnabled>{{config["AccountLevel"]["AdvancedDataProtectionMetrics"]["IsEnabled"]}}</IsEnabled>
        </AdvancedDataProtectionMetrics>
        <DetailedStatusCodesMetrics>
            <IsEnabled>{{config["AccountLevel"]["DetailedStatusCodesMetrics"]["IsEnabled"]}}</IsEnabled>
        </DetailedStatusCodesMetrics>
   </AccountLevel>
   <AwsOrg>
        <Arn>{{config.get("AwsOrg", {}).get("Arn", "")}}</Arn>
    </AwsOrg>
    <StorageLensArn>{{config.get("StorageLensArn")}}</StorageLensArn>
</StorageLensConfiguration>
"""


LIST_STORAGE_LENS_CONFIGURATIONS_TEMPLATE = """
<ListStorageLensConfigurationsResult>
   {% if next_token %}
   <NextToken>{{ next_token }}</NextToken>
   {% endif %}
   {% for config in configs %}
   <StorageLensConfiguration>
      <HomeRegion></HomeRegion>
      <Id>{{ config.config.get("Id") }}</Id>
      <IsEnabled>{{ config.config.get("IsEnabled") }}</IsEnabled>
      <StorageLensArn>{{ config.arn }}</StorageLensArn>
    </StorageLensConfiguration>
    {% endfor %}
</ListStorageLensConfigurationsResult>
"""


GET_STORAGE_LENS_CONFIGURATION_TAGGING_TEMPLATE = """
<GetStorageLensConfigurationTaggingResult>
   <Tags>
      {% for tag in tags["Tag"] %}
      <Tag>
         <Key>{{ tag["Key"] }}</Key>
         <Value>{{ tag["Value"] }}</Value>
      </Tag>
      {% endfor %}
   </Tags>
</GetStorageLensConfigurationTaggingResult>

"""
LIST_ACCESS_POINTS_TEMPLATE = """<ListAccessPointsResult>
  <AccessPointList>
    {% for access_point in access_points %}
    <AccessPoint>
      <Name>{{ access_point.name }}</Name>
      <NetworkOrigin>{{ access_point.network_origin }}</NetworkOrigin>
      {% if access_point.vpc_id %}
      <VpcConfiguration>
        <VpcId>{{ access_point.vpc_id }}</VpcId>
      </VpcConfiguration>
      {% endif %}
      <Bucket>{{ access_point.bucket }}</Bucket>
      <AccessPointArn>{{ access_point.arn }}</AccessPointArn>
      <Alias>{{ access_point.alias }}</Alias>
    </AccessPoint>
    {% endfor %}
  </AccessPointList>
  {% if next_token %}
  <NextToken>{{ next_token }}</NextToken>
  {% endif %}
</ListAccessPointsResult>"""



"""s3vectors base URL and path."""

from .responses import S3VectorsResponse

url_bases = [
    r"https?://s3vectors\.(.+)\.api\.aws",
]

url_paths = {
    "{0}/CreateVectorBucket$": S3VectorsResponse.dispatch,
    "{0}/DeleteVectorBucket$": S3VectorsResponse.dispatch,
    "{0}/GetVectorBucket$": S3VectorsResponse.dispatch,
    "{0}/ListVectorBuckets$": S3VectorsResponse.dispatch,
    "{0}/CreateIndex$": S3VectorsResponse.dispatch,
    "{0}/DeleteIndex$": S3VectorsResponse.dispatch,
    "{0}/GetIndex$": S3VectorsResponse.dispatch,
    "{0}/ListIndexes$": S3VectorsResponse.dispatch,
    "{0}/DeleteVectorBucketPolicy$": S3VectorsResponse.dispatch,
    "{0}/GetVectorBucketPolicy$": S3VectorsResponse.dispatch,
    "{0}/PutVectorBucketPolicy$": S3VectorsResponse.dispatch,
    "{0}/ListVectors$": S3VectorsResponse.dispatch,
    "{0}/DeleteVectors$": S3VectorsResponse.dispatch,
    "{0}/GetVectors$": S3VectorsResponse.dispatch,
    "{0}/PutVectors$": S3VectorsResponse.dispatch,
}



import random
import uuid
from datetime import datetime, timezone
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.vpclattice.exceptions import (
    ResourceNotFoundException,
    ValidationException,
)


class VPCLatticeService(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        auth_type: str,
        certificate_arn: Optional[str],
        client_token: str,
        custom_domain_name: Optional[str],
        name: str,
        tags: Optional[dict[str, str]],
    ) -> None:
        self.id: str = f"svc-{str(uuid.uuid4())[:17]}"
        self.auth_type: str = auth_type
        self.certificate_arn: str = certificate_arn or ""
        self.client_token: str = client_token
        self.custom_domain_name: str = custom_domain_name or ""
        self.dns_entry: VPCLatticeDNSEntry = VPCLatticeDNSEntry(
            region, self.id, self.custom_domain_name
        )
        self.name: str = name
        self.arn: str = f"arn:aws:vpc-lattice:{region}:{account_id}:service/{self.id}"
        self.status: str = "ACTIVE"
        self.tags: dict[str, str] = tags or {}

    def to_dict(self) -> dict[str, Any]:
        return {
            "arn": self.arn,
            "authType": self.auth_type,
            "certificateArn": self.certificate_arn,
            "customDomainName": self.custom_domain_name,
            "dnsEntry": self.dns_entry.to_dict(),
            "id": self.id,
            "name": self.name,
            "status": self.status,
        }


class VPCLatticeServiceNetwork(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        auth_type: str,
        client_token: str,
        name: str,
        sharing_config: Optional[dict[str, Any]],
        tags: Optional[dict[str, str]],
    ) -> None:
        self.auth_type: str = auth_type
        self.client_token: str = client_token
        self.id: str = f"sn-{str(uuid.uuid4())[:17]}"
        self.name: str = name
        self.arn: str = (
            f"arn:aws:vpc-lattice:{region}:{account_id}:servicenetwork/{self.id}"
        )
        self.sharing_config: dict[str, Any] = sharing_config or {}
        self.tags: dict[str, str] = tags or {}

    def to_dict(self) -> dict[str, Any]:
        return {
            "arn": self.arn,
            "authType": self.auth_type,
            "id": self.id,
            "name": self.name,
            "sharingConfig": self.sharing_config,
        }


class VPCLatticeServiceNetworkVpcAssociation(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        client_token: str,
        security_group_ids: Optional[list[str]],
        service_network_identifier: str,
        tags: Optional[dict[str, str]],
        vpc_identifier: str,
    ) -> None:
        self.id: str = f"snva-{service_network_identifier[:4]}-{vpc_identifier[:4]}"
        self.arn: str = f"arn:aws:vpc-lattice:{region}:{account_id}:servicenetworkvpcassociation/{self.id}"
        self.created_by: str = "user"
        self.security_group_ids: list[str] = security_group_ids or []
        self.status: str = "ACTIVE"
        self.tags: dict[str, str] = tags or {}

    def to_dict(self) -> dict[str, Any]:
        return {
            "arn": self.arn,
            "createdBy": self.created_by,
            "id": self.id,
            "securityGroupIds": self.security_group_ids,
            "status": self.status,
            "tags": self.tags,
        }


class VPCLatticeRule(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        action: dict[str, Any],
        client_token: str,
        listener_identifier: str,
        match: dict[str, Any],
        name: str,
        priority: int,
        service_identifier: str,
        tags: dict[str, str],
    ) -> None:
        self.action: dict[str, Any] = action or {}
        self.id: str = f"rule-[0-9a-z]{17}"
        self.arn: str = (
            f"arn:aws:vpc-lattice:{region}:{account_id}:service/{service_identifier}"
            f"/listener/listener-{listener_identifier}/rule/{self.id}"
        )
        self.client_token: str = client_token
        self.listener_identifier: str = listener_identifier
        self.match: dict[str, Any] = match or {}
        self.name: str = name
        self.priority: int = priority
        self.service_identifier: str = service_identifier
        self.tags: dict[str, str] = tags or {}

    def to_dict(self) -> dict[str, Any]:
        return {
            "arn": self.arn,
            "id": self.id,
            "name": self.name,
            "priority": self.priority,
            "action": self.action,
            "match": self.match,
            "serviceIdentifier": self.service_identifier,
            "listenerIdentifier": self.listener_identifier,
            "tags": self.tags,
        }


class VPCLatticeDNSEntry:
    def __init__(
        self,
        region_name: str,
        service_id: str,
        custom_domain_name: Optional[str] = None,
    ) -> None:
        self.domain_name: str = (
            custom_domain_name or f"{service_id}.{region_name}.vpclattice.amazonaws.com"
        )
        self.hosted_zone_id: str = f"Z{random.randint(100000, 999999)}XYZ"

    def to_dict(self) -> dict[str, str]:
        return {"domainName": self.domain_name, "hostedZoneId": self.hosted_zone_id}


class VPCLatticeAccessLogSubscription(BaseModel):
    def __init__(
        self,
        region: str,
        account_id: str,
        destinationArn: str,
        resourceArn: str,
        resourceId: str,  # resourceIdentifier
        serviceNetworkLogType: Optional[str],
        tags: Optional[dict[str, str]],
    ) -> None:
        self.id: str = f"als-{str(uuid.uuid4())[:17]}"
        self.arn: str = (
            f"arn:aws:vpc-lattice:{region}:{account_id}:accesslogsubscription/{self.id}"
        )
        self.created_at = datetime.now(timezone.utc).isoformat()
        self.destinationArn = destinationArn
        self.last_updated_at = datetime.now(timezone.utc).isoformat()
        self.resourceArn = resourceArn
        self.resourceId = resourceId
        self.serviceNetworkLogType = serviceNetworkLogType or "SERVICE"
        self.tags = tags or {}

    def to_dict(self) -> dict[str, Any]:
        return {
            "arn": self.arn,
            "createdAt": self.created_at,
            "destinationArn": self.destinationArn,
            "id": self.id,
            "lastUpdatedAt": self.last_updated_at,
            "resourceArn": self.resourceArn,
            "resourceId": self.resourceId,
            "serviceNetworkLogType": self.serviceNetworkLogType,
            "tags": self.tags,
        }


class AuthPolicy:
    def __init__(
        self,
        policy_name: str,
        policy: str,
        created_at: str,
        last_updated_at: str,
        state: str,
    ) -> None:
        self.policy_name = policy_name
        self.policy = policy
        self.created_at = created_at
        self.last_updated_at = last_updated_at
        self.state = state


class VPCLatticeBackend(BaseBackend):
    PAGINATION_MODEL = {
        "list_services": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 50,
            "unique_attribute": "id",
        },
        "list_service_networks": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 50,
            "unique_attribute": "id",
        },
    }

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.services: dict[str, VPCLatticeService] = {}
        self.service_networks: dict[str, VPCLatticeServiceNetwork] = {}
        self.service_network_vpc_associations: dict[
            str, VPCLatticeServiceNetworkVpcAssociation
        ] = {}
        self.rules: dict[str, VPCLatticeRule] = {}
        self.tagger: TaggingService = TaggingService()
        self.access_log_subscriptions: dict[str, VPCLatticeAccessLogSubscription] = {}
        self.resource_policies: dict[str, str] = {}
        self.auth_policies: dict[str, AuthPolicy] = {}

    def create_service(
        self,
        auth_type: str,
        certificate_arn: Optional[str],
        client_token: str,
        custom_domain_name: Optional[str],
        name: str,
        tags: Optional[dict[str, str]],
    ) -> VPCLatticeService:
        service = VPCLatticeService(
            self.region_name,
            self.account_id,
            auth_type,
            certificate_arn,
            client_token,
            custom_domain_name,
            name,
            tags,
        )
        self.services[service.id] = service
        self.tag_resource(service.arn, tags or {})
        return service

    def get_service(self, service_identifier: str) -> VPCLatticeService:
        service = self.services.get(service_identifier)
        if not service:
            raise ResourceNotFoundException(service_identifier)
        return service

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_services(self) -> list[VPCLatticeService]:
        return list(self.services.values())

    def create_service_network(
        self,
        auth_type: str,
        client_token: str,
        name: str,
        sharing_config: Optional[dict[str, Any]],
        tags: Optional[dict[str, str]],
    ) -> VPCLatticeServiceNetwork:
        """
        WARNING: This method currently does NOT fail if there is a disassociation in progress.
        """
        sn = VPCLatticeServiceNetwork(
            self.region_name,
            self.account_id,
            auth_type,
            client_token,
            name,
            sharing_config,
            tags,
        )
        self.service_networks[sn.id] = sn
        self.tag_resource(sn.arn, tags or {})
        return sn

    def get_service_network(
        self, service_network_identifier: str
    ) -> VPCLatticeServiceNetwork:
        service_network = self.service_networks.get(service_network_identifier)
        if not service_network:
            raise ResourceNotFoundException(service_network_identifier)
        return service_network

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_service_networks(self) -> list[VPCLatticeServiceNetwork]:
        return list(self.service_networks.values())

    def create_service_network_vpc_association(
        self,
        client_token: str,
        security_group_ids: Optional[list[str]],
        service_network_identifier: str,
        tags: Optional[dict[str, str]],
        vpc_identifier: str,
    ) -> VPCLatticeServiceNetworkVpcAssociation:
        assoc = VPCLatticeServiceNetworkVpcAssociation(
            self.region_name,
            self.account_id,
            client_token,
            security_group_ids,
            service_network_identifier,
            tags,
            vpc_identifier,
        )
        self.service_network_vpc_associations[assoc.id] = assoc
        self.tag_resource(assoc.arn, tags or {})
        return assoc

    def create_rule(
        self,
        action: dict[str, Any],
        client_token: str,
        listener_identifier: str,
        match: dict[str, Any],
        name: str,
        priority: int,
        service_identifier: str,
        tags: dict[str, str],
    ) -> VPCLatticeRule:
        rule = VPCLatticeRule(
            self.region_name,
            self.account_id,
            action,
            client_token,
            listener_identifier,
            match,
            name,
            priority,
            service_identifier,
            tags,
        )
        self.rules[rule.id] = rule
        self.tag_resource(rule.arn, tags or {})
        return rule

    def tag_resource(self, resource_arn: str, tags: dict[str, str]) -> None:
        tags_input = self.tagger.convert_dict_to_tags_input(tags or {})
        self.tagger.tag_resource(resource_arn, tags_input)

    def list_tags_for_resource(self, resource_arn: str) -> dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        if not isinstance(tag_keys, list):
            tag_keys = [tag_keys]
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def create_access_log_subscription(
        self,
        resourceIdentifier: str,
        destinationArn: str,
        client_token: Optional[str],
        serviceNetworkLogType: Optional[str],
        tags: Optional[dict[str, str]],
    ) -> VPCLatticeAccessLogSubscription:
        resource: Any = None
        if resourceIdentifier.startswith("sn-"):
            resource = self.service_networks.get(resourceIdentifier)
        elif resourceIdentifier.startswith("svc-"):
            resource = self.services.get(resourceIdentifier)
        else:
            raise ValidationException(
                "Invalid parameter resourceIdentifier, must start with 'sn-' or 'svc-'"
            )

        if not resource:
            raise ResourceNotFoundException(f"Resource {resourceIdentifier} not found")

        sub = VPCLatticeAccessLogSubscription(
            self.region_name,
            self.account_id,
            destinationArn,
            resource.arn,
            resource.id,
            serviceNetworkLogType,
            tags,
        )

        self.access_log_subscriptions[sub.id] = sub
        return sub

    def get_access_log_subscription(
        self, accessLogSubscriptionIdentifier: str
    ) -> VPCLatticeAccessLogSubscription:
        sub = self.access_log_subscriptions.get(accessLogSubscriptionIdentifier)
        if not sub:
            raise ResourceNotFoundException(
                f"Access Log Subscription {accessLogSubscriptionIdentifier} not found"
            )
        return sub

    def list_access_log_subscriptions(
        self,
        resourceIdentifier: str,
        maxResults: Optional[int] = None,
        nextToken: Optional[str] = None,
    ) -> list[VPCLatticeAccessLogSubscription]:
        return [
            sub
            for sub in self.access_log_subscriptions.values()
            if sub.resourceId == resourceIdentifier
            or sub.resourceArn == resourceIdentifier
        ][:maxResults]

    def update_access_log_subscription(
        self,
        accessLogSubscriptionIdentifier: str,
        destinationArn: str,
    ) -> VPCLatticeAccessLogSubscription:
        sub = self.access_log_subscriptions.get(accessLogSubscriptionIdentifier)
        if not sub:
            raise ResourceNotFoundException(
                f"Access Log Subscription {accessLogSubscriptionIdentifier} not found"
            )

        sub.destinationArn = destinationArn
        sub.last_updated_at = datetime.now(timezone.utc).isoformat()

        return sub

    def delete_access_log_subscription(
        self, accessLogSubscriptionIdentifier: str
    ) -> None:
        sub = self.access_log_subscriptions.get(accessLogSubscriptionIdentifier)
        if not sub:
            raise ResourceNotFoundException(
                f"Access Log Subscription {accessLogSubscriptionIdentifier} not found"
            )
        del self.access_log_subscriptions[accessLogSubscriptionIdentifier]

    def put_auth_policy(self, resourceIdentifier: str, policy: str) -> AuthPolicy:
        # ResourceConfig not supported yet, only handle Service and Service Network
        if resourceIdentifier.startswith("arn:"):
            resourceIdentifier = resourceIdentifier.rsplit("/", 1)[-1]

        resource: Any = None

        if resourceIdentifier.startswith("sn-"):
            resource = self.service_networks.get(resourceIdentifier)
        elif resourceIdentifier.startswith("svc-"):
            resource = self.services.get(resourceIdentifier)
        else:
            raise ValidationException(
                "Invalid parameter resourceIdentifier, must start with 'sn-' or 'svc-'"
            )

        if not resource:
            raise ResourceNotFoundException(f"Resource {resourceIdentifier} not found")

        # Handle state management
        state = "Inactive" if resource.auth_type == "NONE" else "Active"

        now_iso = datetime.now(timezone.utc).isoformat()
        if resourceIdentifier in self.auth_policies:
            auth_policy = self.auth_policies[resourceIdentifier]
            auth_policy.policy = policy
            auth_policy.last_updated_at = now_iso
            auth_policy.state = state

        else:
            auth_policy = AuthPolicy(
                policy_name=resourceIdentifier,
                policy=policy,
                created_at=now_iso,
                last_updated_at=now_iso,
                state=state,
            )

            self.auth_policies[resourceIdentifier] = auth_policy

        return auth_policy

    def get_auth_policy(self, resourceIdentifier: str) -> AuthPolicy:
        original_identifier = resourceIdentifier
        if resourceIdentifier.startswith("arn:"):
            resourceIdentifier = resourceIdentifier.rsplit("/", 1)[-1]

        auth_policy = self.auth_policies.get(resourceIdentifier)
        if not auth_policy:
            raise ResourceNotFoundException(f"Resource {original_identifier} not found")

        resource: Any
        if resourceIdentifier.startswith("sn-"):
            resource = self.service_networks.get(resourceIdentifier)
        else:
            resource = self.services.get(resourceIdentifier)

        auth_policy.state = (
            "Inactive" if not resource or resource.auth_type == "NONE" else "Active"
        )

        return auth_policy

    def delete_auth_policy(self, resourceIdentifier: str) -> None:
        original_identifier = resourceIdentifier

        if resourceIdentifier.startswith("arn:"):
            resourceIdentifier = resourceIdentifier.rsplit("/", 1)[-1]

        if resourceIdentifier not in self.auth_policies:
            raise ResourceNotFoundException(f"Resource {original_identifier} not found")

        del self.auth_policies[resourceIdentifier]

    def put_resource_policy(self, resourceArn: str, policy: str) -> None:
        self.resource_policies[resourceArn] = policy

    def get_resource_policy(self, resourceArn: str) -> str:
        if resourceArn not in self.resource_policies:
            raise ResourceNotFoundException(f"Resource {resourceArn} not found")

        return self.resource_policies[resourceArn]

    def delete_resource_policy(self, resourceArn: str) -> None:
        if resourceArn not in self.resource_policies:
            raise ResourceNotFoundException(f"Resource {resourceArn} not found")

        del self.resource_policies[resourceArn]


vpclattice_backends: BackendDict[VPCLatticeBackend] = BackendDict(
    VPCLatticeBackend, "vpc-lattice"
)



"""servicecatalogappregistry base URL and path."""

from .responses import AppRegistryResponse

url_bases = [
    r"https?://servicecatalog-appregistry\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/applications$": AppRegistryResponse.dispatch,
    "{0}/applications/(?P<application>[^/]+)/resources/(?P<resource_type>[^/]+)/(?P<resource>.+)$": AppRegistryResponse.dispatch,
    "{0}/applications/(?P<application>.+)/resources$": AppRegistryResponse.dispatch,
    "{0}/configuration$": AppRegistryResponse.dispatch,
}



import json

from moto.core.responses import BaseResponse

from .models import ResourceGroupsTaggingAPIBackend, resourcegroupstaggingapi_backends


class ResourceGroupsTaggingAPIResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="resourcegroupstaggingapi")

    @property
    def backend(self) -> ResourceGroupsTaggingAPIBackend:
        return resourcegroupstaggingapi_backends[self.current_account][self.region]

    def get_resources(self) -> str:
        pagination_token = self._get_param("PaginationToken")
        tag_filters = self._get_param("TagFilters", [])
        resources_per_page = self._get_int_param("ResourcesPerPage", 50)
        tags_per_page = self._get_int_param("TagsPerPage", 100)
        resource_type_filters = self._get_param("ResourceTypeFilters", [])

        pagination_token, resource_tag_mapping_list = self.backend.get_resources(
            pagination_token=pagination_token,
            tag_filters=tag_filters,
            resources_per_page=resources_per_page,
            tags_per_page=tags_per_page,
            resource_type_filters=resource_type_filters,
        )

        # Format tag response
        response = {
            "ResourceTagMappingList": resource_tag_mapping_list,
            "PaginationToken": pagination_token,
        }
        return json.dumps(response)

    def get_tag_keys(self) -> str:
        pagination_token = self._get_param("PaginationToken")
        pagination_token, tag_keys = self.backend.get_tag_keys(
            pagination_token=pagination_token
        )

        response = {"TagKeys": tag_keys, "PaginationToken": pagination_token}
        return json.dumps(response)

    def get_tag_values(self) -> str:
        pagination_token = self._get_param("PaginationToken")
        key = self._get_param("Key")

        pagination_token, tag_values = self.backend.get_tag_values(
            pagination_token=pagination_token, key=key
        )

        response = {"TagValues": tag_values, "PaginationToken": pagination_token}
        return json.dumps(response)

    def tag_resources(self) -> str:
        resource_arns = self._get_param("ResourceARNList")
        tags = self._get_param("Tags")
        failed_resources = self.backend.tag_resources(
            resource_arns=resource_arns, tags=tags
        )

        return json.dumps({"FailedResourcesMap": failed_resources})

    def untag_resources(self) -> str:
        resource_arn_list = self._get_param("ResourceARNList")
        tag_keys = self._get_param("TagKeys")

        failed_resources = self.backend.untag_resources(
            resource_arn_list=resource_arn_list,
            tag_keys=tag_keys,
        )

        return json.dumps({"FailedResourcesMap": failed_resources})



"""Handles incoming servicecatalog requests, invokes methods, returns responses."""

import json

from moto.core.responses import ActionResult, BaseResponse, EmptyResult

from .models import ServiceCatalogBackend, servicecatalog_backends


class ServiceCatalogResponse(BaseResponse):
    """Handler for ServiceCatalog requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="servicecatalog")

    @property
    def servicecatalog_backend(self) -> ServiceCatalogBackend:
        """Return backend instance specific for this region."""
        # TODO
        # servicecatalog_backends is not yet typed
        # Please modify moto/backends.py to add the appropriate type annotations for this service
        return servicecatalog_backends[self.current_account][self.region]

    def list_portfolio_access(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        portfolio_id = params.get("PortfolioId")
        organization_parent_id = params.get("OrganizationParentId")
        page_token = params.get("PageToken")
        page_size = params.get("PageSize")

        account_id_objects, next_page_token = (
            self.servicecatalog_backend.list_portfolio_access(
                accept_language=accept_language,
                portfolio_id=portfolio_id,
                organization_parent_id=organization_parent_id,
                page_token=page_token,
                page_size=page_size,
            )
        )

        account_ids = [obj["account_id"] for obj in account_id_objects]

        return ActionResult(
            {"AccountIds": account_ids, "NextPageToken": next_page_token}
        )

    def delete_portfolio(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        id = params.get("Id")

        self.servicecatalog_backend.delete_portfolio(
            accept_language=accept_language,
            id=id,
        )

        return EmptyResult()

    def delete_portfolio_share(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        portfolio_id = params.get("PortfolioId")
        account_id = params.get("AccountId")
        organization_node = params.get("OrganizationNode")

        portfolio_share_token = self.servicecatalog_backend.delete_portfolio_share(
            accept_language=accept_language,
            portfolio_id=portfolio_id,
            account_id=account_id,
            organization_node=organization_node,
        )

        response = {}
        if portfolio_share_token:
            response["PortfolioShareToken"] = portfolio_share_token

        return ActionResult(response)

    def create_portfolio(self) -> ActionResult:
        params = json.loads(self.body)

        accept_language = params.get("AcceptLanguage")
        display_name = params.get("DisplayName")
        description = params.get("Description")
        provider_name = params.get("ProviderName")
        tags = params.get("Tags")
        idempotency_token = params.get("IdempotencyToken")

        portfolio_detail, tags = self.servicecatalog_backend.create_portfolio(
            accept_language=accept_language,
            display_name=display_name,
            description=description,
            provider_name=provider_name,
            tags=tags,
            idempotency_token=idempotency_token,
        )

        return ActionResult({"PortfolioDetail": portfolio_detail, "Tags": tags})

    def create_portfolio_share(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        portfolio_id = params.get("PortfolioId")
        account_id = params.get("AccountId")
        organization_node = params.get("OrganizationNode")
        share_tag_options = params.get("ShareTagOptions", False)
        share_principals = params.get("SharePrincipals", False)

        portfolio_share_token = self.servicecatalog_backend.create_portfolio_share(
            accept_language=accept_language,
            portfolio_id=portfolio_id,
            account_id=account_id,
            organization_node=organization_node,
            share_tag_options=share_tag_options,
            share_principals=share_principals,
        )

        response = {}
        if portfolio_share_token:
            response["PortfolioShareToken"] = portfolio_share_token

        return ActionResult(response)

    def list_portfolios(self) -> ActionResult:
        params = self._get_params()
        accept_language = params.get("AcceptLanguage")
        page_token = params.get("PageToken")
        page_size = params.get("PageSize")

        portfolio_details, next_page_token = (
            self.servicecatalog_backend.list_portfolios(
                accept_language=accept_language,
                page_token=page_token,
                page_size=page_size,
            )
        )

        response = {
            "PortfolioDetails": portfolio_details,
            "NextPageToken": next_page_token,
        }
        return ActionResult(response)

    def describe_portfolio_shares(self) -> ActionResult:
        params = json.loads(self.body)
        portfolio_id = params.get("PortfolioId")
        type = params.get("Type")
        page_token = params.get("PageToken")
        page_size = params.get("PageSize")

        next_page_token, portfolio_share_details = (
            self.servicecatalog_backend.describe_portfolio_shares(
                portfolio_id=portfolio_id,
                type=type,
                page_token=page_token,
                page_size=page_size,
            )
        )

        response = {
            "NextPageToken": next_page_token,
            "PortfolioShareDetails": portfolio_share_details,
        }

        return ActionResult(response)

    def describe_portfolio(self) -> ActionResult:
        """Handle describe_portfolio request."""
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        id = params.get("Id")

        portfolio_detail, tags, tag_options, budgets = (
            self.servicecatalog_backend.describe_portfolio(
                accept_language=accept_language,
                id=id,
            )
        )

        response = {
            "PortfolioDetail": portfolio_detail,
            "Tags": tags,
            "TagOptions": tag_options,
            "Budgets": budgets,
        }

        return ActionResult(response)

    def create_product(self) -> ActionResult:
        params = json.loads(self.body)
        name = params.get("Name")
        owner = params.get("Owner")
        description = params.get("Description")
        distributor = params.get("Distributor")
        support_description = params.get("SupportDescription")
        support_email = params.get("SupportEmail")
        support_url = params.get("SupportUrl")
        product_type = params.get("ProductType")
        tags = params.get("Tags")
        provisioning_artifact_parameters = params.get("ProvisioningArtifactParameters")
        idempotency_token = params.get("IdempotencyToken")
        source_connection = params.get("SourceConnection")
        accept_language = params.get("AcceptLanguage")

        response = self.servicecatalog_backend.create_product(
            name=name,
            owner=owner,
            description=description,
            distributor=distributor,
            support_description=support_description,
            support_email=support_email,
            support_url=support_url,
            product_type=product_type,
            tags=tags,
            provisioning_artifact_parameters=provisioning_artifact_parameters,
            idempotency_token=idempotency_token,
            source_connection=source_connection,
            accept_language=accept_language,
        )

        product_view = {
            "ProductViewSummary": response.to_dict(),
            "Status": "AVAILABLE",
            "ProductARN": response.arn,
            "CreatedTime": response.created_time,
            "SourceConnection": response.source_connection,
        }

        return ActionResult(
            {
                "ProductViewDetail": product_view,
                "ProvisioningArtifactDetail": {},
                "Tags": response.tags,
            }
        )

    def describe_product(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        id = params.get("Id")
        name = params.get("Name")

        response = self.servicecatalog_backend.describe_product(
            accept_language=accept_language, id=id, name=name
        )

        result = {
            "ProductViewSummary": response.to_dict(),
            "ProvisioningArtifacts": [],
            "Budgets": [],
            "LaunchPaths": [],
        }

        return ActionResult(result)

    def delete_product(self) -> ActionResult:
        params = json.loads(self.body)
        accept_language = params.get("AcceptLanguage")
        id = params.get("Id")

        self.servicecatalog_backend.delete_product(
            accept_language=accept_language,
            id=id,
        )

        return EmptyResult()



import string
from typing import Any

from moto.moto_api._internal import mock_random as random

from .exceptions import MessageAttributesInvalid


def generate_receipt_handle() -> str:
    # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/ImportantIdentifiers.html#ImportantIdentifiers-receipt-handles
    length = 185
    return "".join(random.choice(string.ascii_lowercase) for x in range(length))


def validate_message_attributes(message_attributes: dict[str, Any]) -> None:
    for name, value in (message_attributes or {}).items():
        data_type = value["DataType"]

        if not data_type:
            raise MessageAttributesInvalid(
                f"The message attribute '{name}' must contain non-empty message attribute value."
            )

        data_type_parts = data_type.split(".")[0]
        if data_type_parts not in [
            "String",
            "Binary",
            "Number",
        ]:
            raise MessageAttributesInvalid(
                f"The message attribute '{name}' has an invalid message attribute type, the set of supported type prefixes is Binary, Number, and String."
            )

        possible_value_fields = ["StringValue", "BinaryValue"]
        for field in possible_value_fields:
            if field in value and value[field] is None:
                raise MessageAttributesInvalid(
                    f"The message attribute '{name}' must contain non-empty message attribute value for message attribute type '{data_type}'."
                )



from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class Settings(EC2BaseResponse):
    def disable_ebs_encryption_by_default(self) -> ActionResult:
        self.error_on_dryrun()
        self.ec2_backend.disable_ebs_encryption_by_default()
        return ActionResult({"EbsEncryptionByDefault": False})

    def enable_ebs_encryption_by_default(self) -> ActionResult:
        self.error_on_dryrun()
        self.ec2_backend.enable_ebs_encryption_by_default()
        return ActionResult({"EbsEncryptionByDefault": True})

    def get_ebs_encryption_by_default(self) -> ActionResult:
        self.error_on_dryrun()
        value = self.ec2_backend.get_ebs_encryption_by_default()
        return ActionResult({"EbsEncryptionByDefault": value})



import abc
from typing import Any, Final, Optional

from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payload_value import (
    PayloadValue,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringExpressionSimple,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class PayloadBinding(PayloadValue, abc.ABC):
    field: Final[str]

    def __init__(self, field: str):
        self.field = field

    def _field_name(self) -> Optional[str]:
        return self.field

    @abc.abstractmethod
    def _eval_val(self, env: Environment) -> Any: ...

    def _eval_body(self, env: Environment) -> None:
        cnt: dict = env.stack.pop()
        val = self._eval_val(env=env)
        cnt[self.field] = val
        env.stack.append(cnt)


class PayloadBindingStringExpressionSimple(PayloadBinding):
    string_expression_simple: Final[StringExpressionSimple]

    def __init__(self, field: str, string_expression_simple: StringExpressionSimple):
        super().__init__(field=field)
        self.string_expression_simple = string_expression_simple

    def _field_name(self) -> Optional[str]:
        return f"{self.field}.$"

    def _eval_val(self, env: Environment) -> Any:
        self.string_expression_simple.eval(env=env)
        value = env.stack.pop()
        return value


class PayloadBindingValue(PayloadBinding):
    payload_value: Final[PayloadValue]

    def __init__(self, field: str, payload_value: PayloadValue):
        super().__init__(field=field)
        self.payload_value = payload_value

    def _eval_val(self, env: Environment) -> Any:
        self.payload_value.eval(env)
        val: Any = env.stack.pop()
        return val



import logging
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.parargs import Parameters
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_worker import (
    IterationWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    JobPool,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

LOG = logging.getLogger(__name__)


class InlineItemProcessorWorker(IterationWorker):
    _parameters: Final[Optional[Parameters]]
    _item_selector: Final[Optional[ItemSelector]]

    def __init__(
        self,
        work_name: str,
        job_pool: JobPool,
        env: Environment,
        item_selector: Optional[ItemSelector],
        parameters: Optional[Parameters],
    ):
        super().__init__(work_name=work_name, job_pool=job_pool, env=env)
        self._item_selector = item_selector
        self._parameters = parameters

    def _eval_input(self, env_frame: Environment) -> None:
        if not self._parameters and not self._item_selector:
            return

        map_state_input = self._env.stack[-1]
        env_frame.states.reset(input_value=map_state_input)
        env_frame.stack.append(map_state_input)

        if self._item_selector:
            self._item_selector.eval(env_frame)
        elif self._parameters:
            self._parameters.eval(env_frame)

        output_value = env_frame.stack[-1]
        env_frame.states.reset(input_value=output_value)



"""kinesisanalyticsv2 base URL and path."""

from .responses import KinesisAnalyticsV2Response

url_bases = [
    r"https?://kinesisanalytics\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": KinesisAnalyticsV2Response.dispatch,
}



"""Module for loading various model files along with any `moto-extras`

Heavily leverages the existing botocore loading mechanism, which expects
a particular directory layout and file naming convention:

    <MOTO_ROOT>
      |
      |-- service_name
      |   |-- YYYY-MM-DD (matching API version for service_name)
      |       |-- service-2.moto-extras.json
      |       |-- paginators-1.moto-extras.json

The ``moto-extras`` files represent extra data to be applied to the model
after it is loaded. Data in these files could augment or modify the AWS
service model by supplying additional error shapes or default input
values for use by Moto.
"""

import os

from botocore.loaders import Loader as BotocoreLoader

import moto

moto_root = os.path.dirname(os.path.abspath(moto.__file__))


class Loader(BotocoreLoader):
    BUILTIN_EXTRAS_TYPES = ["moto"]


def create_loader() -> Loader:
    return Loader(extra_search_paths=[moto_root])



from localstack.packages import InstallTarget, Package, PackageInstaller
from localstack.packages.core import MavenPackageInstaller
from localstack.packages.java import JavaInstallerMixin, java_package

JSONATA_DEFAULT_VERSION = "0.9.7"
JACKSON_DEFAULT_VERSION = "2.16.2"

JSONATA_JACKSON_VERSION_STORE = {JSONATA_DEFAULT_VERSION: JACKSON_DEFAULT_VERSION}


class JSONataPackage(Package):
    def __init__(self):
        super().__init__("JSONataLibs", JSONATA_DEFAULT_VERSION)

    def get_versions(self) -> list[str]:
        return list(JSONATA_JACKSON_VERSION_STORE.keys())

    def _get_installer(self, version: str) -> PackageInstaller:
        return JSONataPackageInstaller(version)


class JSONataPackageInstaller(JavaInstallerMixin, MavenPackageInstaller):
    def __init__(self, version: str):
        jackson_version = JSONATA_JACKSON_VERSION_STORE[version]

        # Match the dynamodb-local JRE version to reduce the LocalStack image size by sharing the same JRE version
        self.java_version = "21"

        super().__init__(
            f"pkg:maven/com.dashjoin/jsonata@{version}",
            # jackson-databind is imported in jsonata.py as "from com.fasterxml.jackson.databind import ObjectMapper"
            # jackson-annotations and jackson-core are dependencies of jackson-databind:
            # https://central.sonatype.com/artifact/com.fasterxml.jackson.core/jackson-databind/dependencies
            f"pkg:maven/com.fasterxml.jackson.core/jackson-core@{jackson_version}",
            f"pkg:maven/com.fasterxml.jackson.core/jackson-annotations@{jackson_version}",
            f"pkg:maven/com.fasterxml.jackson.core/jackson-databind@{jackson_version}",
        )

    def _prepare_installation(self, target: InstallTarget) -> None:
        # override to install correct java version
        java_package.get_installer(self.java_version).install(target)

    def get_java_home(self) -> str | None:
        """Override to use the specific Java version"""
        return java_package.get_installer(self.java_version).get_java_home()


jpype_jsonata_package = JSONataPackage()



import json
from typing import Any

from moto.core.responses import BaseResponse

from .models import ResourceAccessManagerBackend, ram_backends


class ResourceAccessManagerResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="ram")

    @property
    def ram_backend(self) -> ResourceAccessManagerBackend:
        return ram_backends[self.current_account][self.region]

    @property
    def request_params(self) -> dict[str, Any]:  # type: ignore[misc]
        try:
            return json.loads(self.body)
        except ValueError:
            return {}

    def create_resource_share(self) -> str:
        return json.dumps(self.ram_backend.create_resource_share(**self.request_params))

    def get_resource_shares(self) -> str:
        resource_owner = self._get_param("resourceOwner")
        return json.dumps(self.ram_backend.get_resource_shares(resource_owner))

    def update_resource_share(self) -> str:
        resource_share_arn = self._get_param("resourceShareArn")
        allow_external_principals = self._get_param("allowExternalPrincipals", True)
        name = self._get_param("name")
        return json.dumps(
            self.ram_backend.update_resource_share(
                resource_share_arn, allow_external_principals, name
            )
        )

    def delete_resource_share(self) -> str:
        resource_share_arn = self._get_param("resourceShareArn")
        return json.dumps(self.ram_backend.delete_resource_share(resource_share_arn))

    def enable_sharing_with_aws_organization(self) -> str:
        return json.dumps(self.ram_backend.enable_sharing_with_aws_organization())

    def get_resource_share_associations(self) -> str:
        association_type = self._get_param("associationType")
        association_status = self._get_param("associationStatus")
        resource_share_arns = self._get_param("resourceShareArns", [])
        resource_arn = self._get_param("resourceArn")
        principal = self._get_param("principal")
        return json.dumps(
            self.ram_backend.get_resource_share_associations(
                association_type,
                association_status,
                resource_share_arns,
                resource_arn,
                principal,
            )
        )

    def list_resource_types(self) -> str:
        resource_region_scope = self._get_param("resourceRegionScope", "ALL")
        return json.dumps(self.ram_backend.list_resource_types(resource_region_scope))

    def list_permissions(self) -> str:
        resource_type = self._get_param("resourceType")
        permission_type = self._get_param("permissionType", "ALL")
        return json.dumps(
            self.ram_backend.list_permissions(resource_type, permission_type)
        )



from typing import Any

from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class ReservedInstances(EC2BaseResponse):
    def cancel_reserved_instances_listing(self) -> None:
        self.error_on_dryrun()

        raise NotImplementedError(
            "ReservedInstances.cancel_reserved_instances_listing is not yet implemented"
        )

    def create_reserved_instances_listing(self) -> None:
        self.error_on_dryrun()

        raise NotImplementedError(
            "ReservedInstances.create_reserved_instances_listing is not yet implemented"
        )

    def describe_reserved_instances(self) -> None:
        raise NotImplementedError(
            "ReservedInstances.describe_reserved_instances is not yet implemented"
        )

    def describe_reserved_instances_listings(self) -> None:
        raise NotImplementedError(
            "ReservedInstances.describe_reserved_instances_listings is not yet implemented"
        )

    def describe_reserved_instances_offerings(self) -> ActionResult:
        self.error_on_dryrun()

        # Basic support for filters commonly used by clients
        # Supported filters:
        #  - availability-zone
        #  - instance-type
        #  - product-description
        #  - offering-type
        filters = self._filters_from_querystring()

        offerings = self.ec2_backend.describe_reserved_instances_offerings(filters)
        result: dict[str, Any] = {"ReservedInstancesOfferings": offerings}
        # Pagination is not implemented; always return the full list
        return ActionResult(result)

    def purchase_reserved_instances_offering(self) -> None:
        self.error_on_dryrun()

        raise NotImplementedError(
            "ReservedInstances.purchase_reserved_instances_offering is not yet implemented"
        )



"""CloudDirectoryBackend class with methods for supported APIs."""

import datetime

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

from .exceptions import (
    InvalidArnException,
    ResourceNotFoundException,
    SchemaAlreadyPublishedException,
)

PAGINATION_MODEL = {
    "list_directories": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "directory_arn",
    },
    "list_development_schema_arns": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "schema_arn",
    },
    "list_published_schema_arns": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "schema_arn",
    },
}


class Directory(BaseModel):
    def __init__(
        self, account_id: str, region: str, name: str, schema_arn: str
    ) -> None:
        self.name = name
        self.schema_arn = schema_arn
        self.directory_arn = (
            f"arn:aws:clouddirectory:{region}:{account_id}:directory/{name}"
        )
        self.state = "ENABLED"
        self.creation_date_time = datetime.datetime.now()
        self.object_identifier = f"directory-{name}"

    def to_dict(self) -> dict[str, str]:
        return {
            "Name": self.name,
            "SchemaArn": self.schema_arn,
            "DirectoryArn": self.directory_arn,
            "State": self.state,
            "CreationDateTime": str(self.creation_date_time),
            "ObjectIdentifier": self.object_identifier,
        }


class CloudDirectoryBackend(BaseBackend):
    """Implementation of CloudDirectory APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.directories: dict[str, Directory] = {}
        self.schemas_states: dict[str, list[str]] = {
            "development": [],
            "published": [],
            "applied": [],
        }
        self.tagger = TaggingService()

    def apply_schema(self, directory_arn: str, published_schema_arn: str) -> None:
        directory = self.directories.get(directory_arn)
        if not directory:
            raise ResourceNotFoundException(directory_arn)
        if published_schema_arn not in self.schemas_states["published"]:
            raise ResourceNotFoundException(published_schema_arn)
        directory.schema_arn = published_schema_arn
        return

    def publish_schema(
        self, name: str, version: str, development_schema_arn: str, minor_version: str
    ) -> str:
        schema_arn = f"arn:aws:clouddirectory:{self.region_name}:{self.account_id}:schema/published/{name}/{version}/{minor_version}"
        if development_schema_arn in self.schemas_states["published"]:
            raise SchemaAlreadyPublishedException(development_schema_arn)
        if development_schema_arn in self.schemas_states["development"]:
            self.schemas_states["development"].remove(development_schema_arn)
            self.schemas_states["published"].append(schema_arn)
        else:
            raise ResourceNotFoundException(development_schema_arn)
        return schema_arn

    def create_directory(self, name: str, schema_arn: str) -> Directory:
        directory = Directory(self.account_id, self.region_name, name, schema_arn)
        self.directories[directory.directory_arn] = directory
        return directory

    def create_schema(self, name: str) -> str:
        self.schema_arn = f"arn:aws:clouddirectory:{self.region_name}:{self.account_id}:schema/development/{name}"
        self.schemas_states["development"].append(self.schema_arn)
        return self.schema_arn

    def delete_schema(self, schema_arn: str) -> None:
        if schema_arn in self.schemas_states["development"]:
            self.schemas_states["development"].remove(schema_arn)
        elif schema_arn in self.schemas_states["published"]:
            self.schemas_states["published"].remove(schema_arn)
        else:
            raise ResourceNotFoundException(schema_arn)
        return

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_development_schema_arns(self) -> list[str]:
        return self.schemas_states["development"]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_published_schema_arns(self) -> list[str]:
        return self.schemas_states["published"]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_directories(self, state: str) -> list[Directory]:
        directories = list(self.directories.values())
        if state:
            directories = [
                directory for directory in directories if directory.state == state
            ]
        return directories

    def tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        self.tagger.tag_resource(resource_arn, tags)
        return

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        if not isinstance(tag_keys, list):
            tag_keys = [tag_keys]
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)
        return

    def delete_directory(self, directory_arn: str) -> str:
        directory = self.directories.pop(directory_arn)
        return directory.directory_arn

    def get_directory(self, directory_arn: str) -> Directory:
        directory = self.directories.get(directory_arn)
        if not directory:
            raise InvalidArnException(directory_arn)
        return directory

    def list_tags_for_resource(
        self, resource_arn: str, next_token: str, max_results: int
    ) -> list[dict[str, str]]:
        tags = self.tagger.list_tags_for_resource(resource_arn)["Tags"]
        return tags


clouddirectory_backends = BackendDict(CloudDirectoryBackend, "clouddirectory")



"""SyntheticsBackend class with methods for supported APIs."""

import uuid
from typing import Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds, utcnow


class Canary(BaseModel):
    """
    Represents a CloudWatch Synthetics Canary resource.
    """

    def __init__(
        self,
        name: str,
        code: dict[str, object],
        artifact_s3_location: str,
        execution_role_arn: str,
        schedule: dict[str, object],
        run_config: dict[str, object],
        success_retention_period_in_days: int,
        failure_retention_period_in_days: int,
        runtime_version: str,
        vpc_config: Optional[dict[str, object]],
        resources_to_replicate_tags: Optional[list[str]],
        provisioned_resource_cleanup: Optional[str],
        browser_configs: Optional[list[dict[str, object]]],
        tags: Optional[dict[str, str]],
        artifact_config: Optional[dict[str, object]],
    ):
        self.name = name
        self.id = str(uuid.uuid4())
        self.code = code
        self.artifact_s3_location = artifact_s3_location
        self.execution_role_arn = execution_role_arn
        self.schedule = schedule
        self.run_config = run_config
        self.success_retention_period_in_days = success_retention_period_in_days
        self.failure_retention_period_in_days = failure_retention_period_in_days
        self.runtime_version = runtime_version
        self.vpc_config = vpc_config
        self.resources_to_replicate_tags = resources_to_replicate_tags
        self.provisioned_resource_cleanup = provisioned_resource_cleanup
        self.browser_configs = browser_configs or []
        self.tags = tags or {}
        self.artifact_config = artifact_config
        self.state = "READY"

        now = utcnow()
        self.timeline = {
            "Created": now,
            "LastModified": now,
            "LastStarted": None,
            "LastStopped": None,
        }

    def to_dict(self) -> dict[str, object]:
        """
        Convert the Canary object to a dictionary representation.
        """
        return {
            "Id": self.id,
            "Name": self.name,
            "Code": self.code,
            "ArtifactS3Location": self.artifact_s3_location,
            "ExecutionRoleArn": self.execution_role_arn,
            "Schedule": self.schedule,
            "RunConfig": self.run_config,
            "SuccessRetentionPeriodInDays": self.success_retention_period_in_days,
            "FailureRetentionPeriodInDays": self.failure_retention_period_in_days,
            "RuntimeVersion": self.runtime_version,
            "VpcConfig": self.vpc_config,
            "ProvisionedResourceCleanup": self.provisioned_resource_cleanup,
            "BrowserConfigs": self.browser_configs,
            "Tags": self.tags,
            "ArtifactConfig": self.artifact_config,
            "Status": {
                "State": self.state,
                "StateReason": "Created by Moto",
                "StateReasonCode": "CREATE_COMPLETE",
            },
            "Timeline": {
                "Created": iso_8601_datetime_with_milliseconds(
                    self.timeline["Created"]
                ),
                "LastModified": iso_8601_datetime_with_milliseconds(
                    self.timeline["LastModified"]
                ),
                "LastStarted": None,
                "LastStopped": None,
            },
        }


class SyntheticsBackend(BaseBackend):
    """Implementation of Synthetics APIs."""

    def __init__(self, region_name: str, account_id: str):
        """
        Initialize the SyntheticsBackend with region and account.
        """
        super().__init__(region_name, account_id)
        self.canaries: dict[str, Canary] = {}

    def create_canary(
        self,
        name: str,
        code: dict[str, object],
        artifact_s3_location: str,
        execution_role_arn: str,
        schedule: dict[str, object],
        run_config: dict[str, object],
        success_retention_period_in_days: int,
        failure_retention_period_in_days: int,
        runtime_version: str,
        vpc_config: Optional[dict[str, object]],
        resources_to_replicate_tags: Optional[list[str]],
        provisioned_resource_cleanup: Optional[str],
        browser_configs: Optional[list[dict[str, object]]],
        tags: Optional[dict[str, str]],
        artifact_config: Optional[dict[str, object]],
    ) -> Canary:
        canary = Canary(
            name,
            code,
            artifact_s3_location,
            execution_role_arn,
            schedule,
            run_config,
            success_retention_period_in_days,
            failure_retention_period_in_days,
            runtime_version,
            vpc_config,
            resources_to_replicate_tags,
            provisioned_resource_cleanup,
            browser_configs,
            tags,
            artifact_config,
        )
        self.canaries[name] = canary
        return canary

    def get_canary(self, name: str, dry_run_id: Optional[str] = None) -> Canary:
        """
        The dry-run_id-parameter is not yet supported
        """
        # dry_run_id is unused, included for API compatibility
        return self.canaries[name]

    def describe_canaries(
        self,
        next_token: Optional[str],
        max_results: Optional[int],
        names: Optional[list[str]],
    ) -> tuple[list[Canary], None]:
        """
        Pagination is not yet supported
        """
        canaries = list(self.canaries.values())
        if names:
            canaries = [c for c in canaries if c.name in names]

        return canaries, None

    def list_tags_for_resource(self, resource_arn: str) -> dict[str, str]:
        # Simplified: assume resource_arn is actually the canary name
        canary = self.canaries.get(resource_arn)
        return canary.tags if canary else {}


# Exported backend dict for Moto Synthetics
synthetics_backends = BackendDict(SyntheticsBackend, "synthetics")



import abc
from typing import Final

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.intrinsic.argument.argument import (
    ArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.function_name import (
    FunctionName,
)


class Function(EvalComponent, abc.ABC):
    name: FunctionName
    argument_list: Final[ArgumentList]

    def __init__(self, name: FunctionName, argument_list: ArgumentList):
        self.name = name
        self.argument_list = argument_list



from moto.core.responses import ActionResult, BaseResponse, EmptyResult

from .exceptions import InvalidParameterValueError
from .models import EBBackend, eb_backends

# AWS Solution Stack Details as of 2025-07-13
SOLUTION_STACK_DETAILS = [
    {
        "SolutionStackName": "64bit Windows Server 2019 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server Core 2019 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server 2025 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server Core 2016 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server Core 2025 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server 2022 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server 2016 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Windows Server Core 2022 v2.19.2 running IIS 10.0",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2017.03 v2.5.4 running Java 8",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 9 Corretto 17",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 11 Corretto 17",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 10 Corretto 21",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 11 Corretto 21",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 10 Corretto 17",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v5.7.0 running Tomcat 9 Corretto 11",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Python 3.11",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.7.0 running PHP 8.2",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.7.0 running PHP 8.4",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Python 3.9",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Python 3.12",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Ruby 3.3",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Ruby 3.4",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Python 3.13",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.7.0 running PHP 8.3",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Ruby 3.2",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v6.6.0 running Node.js 22",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.7.0 running PHP 8.1",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v6.6.0 running Node.js 20",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Corretto 17",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v3.5.0 running .NET 9",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v6.6.0 running Node.js 18",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Corretto 8",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v3.5.0 running .NET 8",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Corretto 11",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Docker",
        "PermittedFileTypes": [],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.2.0 running ECS",
        "PermittedFileTypes": [],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.4.0 running Go 1",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2023 v4.6.0 running Corretto 21",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v4.9.0 running Tomcat 9 Corretto 11",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v2.11.0 running .NET Core",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.9.0 running Corretto 11",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.9.0 running Corretto 8",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.13.0 running Go 1",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v5.11.0 running Node.js 18",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.5.0 running ECS",
        "PermittedFileTypes": [],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.9.0 running Corretto 17",
        "PermittedFileTypes": ["jar", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.10.0 running PHP 8.1",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v4.9.0 running Tomcat 9 Corretto 8",
        "PermittedFileTypes": ["war", "zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v4.2.0 running Docker",
        "PermittedFileTypes": [],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v3.3.14 running Python 3.8",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v5.0.2 running Node.js 12",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v5.0.2 running Node.js 10",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2 v0.1.0 running Node.js 12 (BETA)",
        "PermittedFileTypes": ["zip"],
    },
    {
        "SolutionStackName": "64bit Amazon Linux 2018.03 v2.6.33 running Packer 1.0.3",
        "PermittedFileTypes": [],
    },
]


class EBResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="elasticbeanstalk")
        self.automated_parameter_parsing = True

    @property
    def elasticbeanstalk_backend(self) -> EBBackend:
        """
        :rtype: EBBackend
        """
        return eb_backends[self.current_account][self.region]

    def create_application(self) -> ActionResult:
        app = self.elasticbeanstalk_backend.create_application(
            application_name=self._get_param("ApplicationName")
        )

        result = {"Application": app}
        return ActionResult(result)

    def describe_applications(self) -> ActionResult:
        applications = self.elasticbeanstalk_backend.applications.values()
        result = {"Applications": applications}
        return ActionResult(result)

    def create_environment(self) -> ActionResult:
        application_name = self._get_param("ApplicationName")
        try:
            app = self.elasticbeanstalk_backend.applications[application_name]
        except KeyError:
            raise InvalidParameterValueError(
                f"No Application named '{application_name}' found."
            )

        tags = {tag["Key"]: tag["Value"] for tag in self._get_param("Tags", [])}
        env = self.elasticbeanstalk_backend.create_environment(
            app,
            environment_name=self._get_param("EnvironmentName"),
            stack_name=self._get_param("SolutionStackName"),
            tags=tags,
        )

        result = env
        return ActionResult(result)

    def describe_environments(self) -> ActionResult:
        envs = self.elasticbeanstalk_backend.describe_environments()

        result = {"Environments": envs}
        return ActionResult(result)

    def list_available_solution_stacks(self) -> ActionResult:
        solution_stacks = {
            "SolutionStacks": [
                stack["SolutionStackName"] for stack in SOLUTION_STACK_DETAILS
            ],
            "SolutionStackDetails": SOLUTION_STACK_DETAILS,
        }
        return ActionResult(solution_stacks)

    def update_tags_for_resource(self) -> ActionResult:
        resource_arn = self._get_param("ResourceArn")
        tags_to_add = {
            tag["Key"]: tag["Value"] for tag in self._get_param("TagsToAdd", [])
        }
        tags_to_remove = self._get_param("TagsToRemove", [])
        self.elasticbeanstalk_backend.update_tags_for_resource(
            resource_arn, tags_to_add, tags_to_remove
        )
        return EmptyResult()

    def list_tags_for_resource(self) -> ActionResult:
        resource_arn = self._get_param("ResourceArn")
        tags = self.elasticbeanstalk_backend.list_tags_for_resource(resource_arn)
        result = {
            "ResourceArn": resource_arn,
            "ResourceTags": [{"Key": k, "Value": v} for k, v in tags.items()],
        }
        return ActionResult(result)

    def delete_application(self) -> ActionResult:
        application_name = self._get_param("ApplicationName")
        self.elasticbeanstalk_backend.delete_application(
            application_name=application_name,
        )
        return EmptyResult()



"""Handles incoming networkfirewall requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import NetworkFirewallBackend, networkfirewall_backends


class NetworkFirewallResponse(BaseResponse):
    """Handler for NetworkFirewall requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="networkfirewall")

    @property
    def networkfirewall_backend(self) -> NetworkFirewallBackend:
        """Return backend instance specific for this region."""
        return networkfirewall_backends[self.current_account][self.region]

    def create_firewall(self) -> str:
        firewall_name = self._get_param("FirewallName")
        firewall_policy_arn = self._get_param("FirewallPolicyArn")
        vpc_id = self._get_param("VpcId")
        subnet_mappings = self._get_param("SubnetMappings")
        delete_protection = self._get_param("DeleteProtection")
        subnet_change_protection = self._get_param("SubnetChangeProtection")
        firewall_policy_change_protection = self._get_param(
            "FirewallPolicyChangeProtection"
        )
        description = self._get_param("Description")
        tags = self._get_param("Tags")
        encryption_configuration = self._get_param("EncryptionConfiguration")
        enabled_analysis_types = self._get_param("EnabledAnalysisTypes")
        firewall = self.networkfirewall_backend.create_firewall(
            firewall_name=firewall_name,
            firewall_policy_arn=firewall_policy_arn,
            vpc_id=vpc_id,
            subnet_mappings=subnet_mappings,
            delete_protection=delete_protection,
            subnet_change_protection=subnet_change_protection,
            firewall_policy_change_protection=firewall_policy_change_protection,
            description=description,
            tags=tags,
            encryption_configuration=encryption_configuration,
            enabled_analysis_types=enabled_analysis_types,
        )

        return json.dumps(
            {"Firewall": firewall.to_dict(), "FirewallStatus": firewall.firewall_status}
        )

    def describe_logging_configuration(self) -> str:
        firewall_arn = self._get_param("FirewallArn")
        firewall_name = self._get_param("FirewallName")
        firewall = self.networkfirewall_backend.describe_logging_configuration(
            firewall_arn=firewall_arn,
            firewall_name=firewall_name,
        )
        return json.dumps(
            {
                "FirewallArn": firewall.arn,
                "LoggingConfiguration": firewall.logging_configs,
            }
        )

    def update_logging_configuration(self) -> str:
        firewall_arn = self._get_param("FirewallArn")
        firewall_name = self._get_param("FirewallName")
        logging_configuration = self._get_param("LoggingConfiguration")
        firewall = self.networkfirewall_backend.update_logging_configuration(
            firewall_arn=firewall_arn,
            firewall_name=firewall_name,
            logging_configuration=logging_configuration,
        )
        return json.dumps(
            {
                "FirewallArn": firewall.arn,
                "FirewallName": firewall.firewall_name,
                "LoggingConfiguration": firewall.logging_configs,
            }
        )

    def list_firewalls(self) -> str:
        next_token = self._get_param("NextToken")
        vpc_ids = self._get_param("VpcIds")
        max_results = self._get_param("MaxResults")
        firewalls, next_token = self.networkfirewall_backend.list_firewalls(
            next_token=next_token,
            vpc_ids=vpc_ids,
            max_results=max_results,
        )
        firewall_list = [fw.to_dict() for fw in firewalls]
        return json.dumps({"nextToken": next_token, "Firewalls": firewall_list})

    def describe_firewall(self) -> str:
        firewall_name = self._get_param("FirewallName")
        firewall_arn = self._get_param("FirewallArn")
        firewall = self.networkfirewall_backend.describe_firewall(
            firewall_name=firewall_name,
            firewall_arn=firewall_arn,
        )
        return json.dumps(
            {
                "UpdateToken": firewall.update_token,
                "Firewall": firewall.to_dict(),
                "FirewallStatus": firewall.firewall_status,
            }
        )



from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)


class InvalidParametersException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidParametersException", message)



from moto.core.exceptions import ServiceException


class LifecyclePolicyNotFoundException(ServiceException):
    code = "LifecyclePolicyNotFoundException"

    def __init__(self, repository_name: str, registry_id: str):
        message = (
            f"Lifecycle policy does not exist "
            f"for the repository with name '{repository_name}' "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class LimitExceededException(ServiceException):
    code = "LimitExceededException"
    message = "The scan quota per image has been exceeded. Wait and try again."


class RegistryPolicyNotFoundException(ServiceException):
    code = "RegistryPolicyNotFoundException"

    def __init__(self, registry_id: str):
        message = (
            f"Registry policy does not exist in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class RepositoryAlreadyExistsException(ServiceException):
    code = "RepositoryAlreadyExistsException"

    def __init__(self, repository_name: str, registry_id: str):
        message = (
            f"The repository with name '{repository_name}' already exists "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class RepositoryNotEmptyException(ServiceException):
    code = "RepositoryNotEmptyException"

    def __init__(self, repository_name: str, registry_id: str):
        message = (
            f"The repository with name '{repository_name}' "
            f"in registry with id '{registry_id}' "
            f"cannot be deleted because it still contains images"
        )
        super().__init__(message)


class RepositoryNotFoundException(ServiceException):
    code = "RepositoryNotFoundException"

    def __init__(self, repository_name: str, registry_id: str):
        message = (
            f"The repository with name '{repository_name}' does not exist "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class RepositoryPolicyNotFoundException(ServiceException):
    code = "RepositoryPolicyNotFoundException"

    def __init__(self, repository_name: str, registry_id: str):
        message = (
            f"Repository policy does not exist "
            f"for the repository with name '{repository_name}' "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class ImageNotFoundException(ServiceException):
    code = "ImageNotFoundException"

    def __init__(self, image_id: str, repository_name: str, registry_id: str):
        message = (
            f"The image with imageId {image_id} does not exist "
            f"within the repository with name '{repository_name}' "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class ImageAlreadyExistsException(ServiceException):
    code = "ImageAlreadyExistsException"

    def __init__(
        self,
        repository_name: str,
        registry_id: str,
        digest: str,
        image_tag: str,
    ):
        message = (
            f"Image with digest '{digest}' and tag '{image_tag}' already exists "
            f"in the repository with name '{repository_name}' "
            f"in registry with id '{registry_id}'"
        )
        super().__init__(message)


class InvalidParameterException(ServiceException):
    code = "InvalidParameterException"


class ScanNotFoundException(ServiceException):
    code = "ScanNotFoundException"

    def __init__(self, image_id: str, repository_name: str, registry_id: str):
        message = (
            f"Image scan does not exist for the image with '{image_id}' "
            f"in the repository with name '{repository_name}' "
            f"in the registry with id '{registry_id}'"
        )
        super().__init__(message)


class ValidationException(ServiceException):
    code = "ValidationException"



from __future__ import annotations

import ipaddress
import itertools
from collections import defaultdict
from collections.abc import Iterator
from functools import cache
from typing import TYPE_CHECKING, Any, Literal, Optional, cast

from moto.core.common_models import CloudFormationModel

if TYPE_CHECKING:
    from moto.ec2.models.instances import Instance

from moto.ec2.models.availability_zones_and_regions import Zone

from ..exceptions import (
    DefaultSubnetAlreadyExistsInAvailabilityZoneError,
    DefaultVpcDoesNotExistError,
    InvalidAvailabilityZoneError,
    InvalidCIDRBlockParameterError,
    InvalidCidrReservationNotFound,
    InvalidCidrReservationNotWithinSubnetCidr,
    InvalidCidrReservationOverlapExisting,
    InvalidParameterValueError,
    InvalidPrefixReservationType,
    InvalidSubnetCidrBlockAssociationID,
    InvalidSubnetConflictError,
    InvalidSubnetIdError,
    InvalidSubnetRangeError,
)
from ..utils import (
    generic_filter,
    random_subnet_cidr_reservation_id,
    random_subnet_id,
    random_subnet_ipv6_cidr_block_association_id,
)
from .availability_zones_and_regions import RegionsAndZonesBackend
from .core import TaggedEC2Resource


class SubnetCidrReservation(TaggedEC2Resource):
    def __init__(
        self,
        ec2_backend: Any,
        subnet_id: str,
        cidr: ipaddress.IPv4Network | ipaddress.IPv6Network,
        reservation_type: Literal["explicit"] | Literal["prefix"],
        owner_id: str,
        tag_specifications: Optional[list[dict[str, Any]]] = None,
        subnet_cidr_reservation_id: Optional[str] = None,
    ) -> None:
        self.ec2_backend = ec2_backend
        self.id = subnet_cidr_reservation_id or random_subnet_cidr_reservation_id()
        self.subnet_id = subnet_id
        self.cidr = cidr
        self.reservation_type = reservation_type
        self.owner_id = owner_id
        self.tag_specifications = tag_specifications or []

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> Any:
        if filter_name == "reservationType":
            return self.reservation_type
        elif filter_name == "subnet-id":
            return self.subnet_id

        return super().get_filter_value(filter_name, method_name)


class Subnet(TaggedEC2Resource, CloudFormationModel):
    def __init__(
        self,
        ec2_backend: Any,
        subnet_id: str,
        vpc_id: str,
        cidr_block: str,
        ipv6_cidr_block: Optional[str],
        availability_zone: Zone,
        default_for_az: bool,
        map_public_ip_on_launch: bool,
        ipv6_native: bool = False,
    ):
        self.ec2_backend = ec2_backend
        self.id = subnet_id
        self.vpc_id = vpc_id
        self.cidr_block = cidr_block
        self.cidr = (
            ipaddress.IPv4Network(str(self.cidr_block), strict=False)
            if self.cidr_block
            else ipaddress.IPv6Network(ipv6_cidr_block)
        )
        self._available_ip_addresses = self.cidr.num_addresses - 5
        self._availability_zone = availability_zone
        self.default_for_az = default_for_az
        self.map_public_ip_on_launch = map_public_ip_on_launch
        self.assign_ipv6_address_on_creation = ipv6_native
        self.ipv6_cidr_block_associations: dict[str, dict[str, Any]] = {}
        if ipv6_cidr_block:
            self.attach_ipv6_cidr_block_associations(ipv6_cidr_block)

        self._reserved_ips: set[ipaddress.IPv4Address | ipaddress.IPv6Address] = set()

        # Reserved by AWS
        self.aws_reserved_ips = {
            next(iter(self._subnet_ip_generator)) for _ in range(0, 3)
        }

        self._unused_ips: set[str] = (
            set()
        )  # if instance is destroyed hold IP here for reuse
        self._subnet_ips: dict[str, Instance] = {}
        self.state = "available"

        # Placeholder for response templates until Ipv6 support implemented.
        self.ipv6_native = ipv6_native

        self._cidr_reservations: list[SubnetCidrReservation] = []

    @property
    def arn(self) -> str:
        return f"arn:aws:ec2:{self.ec2_backend.region_name}:{self.owner_id}:subnet/{self.id}"

    @property
    def ipv6_cidr_block_association_set(self) -> list[dict[str, str]]:
        association_set = [
            {
                "ipv6CidrBlock": association["ipv6CidrBlock"],
                "associationId": association["associationId"],
                "ipv6CidrBlockState": {
                    "state": association["ipv6CidrBlockState"],
                },
            }
            for association in self.ipv6_cidr_block_associations.values()
            if association["ipv6CidrBlockState"]["State"] == "associated"
        ]
        return association_set

    @property
    def owner_id(self) -> str:
        return self.ec2_backend.account_id

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnet.html
        return "AWS::EC2::Subnet"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> Subnet:
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        vpc_id = properties["VpcId"]
        cidr_block = properties["CidrBlock"]
        availability_zone = properties.get("AvailabilityZone")
        ec2_backend = ec2_backends[account_id][region_name]
        subnet = ec2_backend.create_subnet(
            vpc_id=vpc_id, cidr_block=cidr_block, availability_zone=availability_zone
        )
        for tag in properties.get("Tags", []):
            tag_key = tag["Key"]
            tag_value = tag["Value"]
            subnet.add_tag(tag_key, tag_value)

        return subnet

    @classmethod
    def delete_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: dict[str, Any],
        account_id: str,
        region_name: str,
    ) -> None:
        from ..models import ec2_backends

        ec2_backends[account_id][region_name].delete_subnet(resource_name)

    @property
    def available_ip_address_count(self) -> int:
        enis = [
            eni
            for eni in self.ec2_backend.get_all_network_interfaces()
            if eni.subnet.id == self.id
        ]
        addresses_taken = []
        for eni in enis:
            if eni.private_ip_addresses:
                addresses_taken.extend(eni.private_ip_addresses)
        return self._available_ip_addresses - len(addresses_taken)

    @property
    def availability_zone(self) -> str:
        return self._availability_zone.name

    @property
    def availability_zone_id(self) -> str:
        return self._availability_zone.zone_id

    @property
    def physical_resource_id(self) -> str:
        return self.id

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> Any:
        """
        API Version 2014-10-01 defines the following filters for DescribeSubnets:

        * availabilityZone
        * available-ip-address-count
        * cidrBlock
        * defaultForAz
        * state
        * subnet-id
        * tag:key=value
        * tag-key
        * tag-value
        * vpc-id

        Taken from: http://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeSubnets.html
        """
        if filter_name in ("cidr", "cidrBlock", "cidr-block"):
            return self.cidr_block
        elif filter_name in ("vpc-id", "vpcId"):
            return self.vpc_id
        elif filter_name == "subnet-id":
            return self.id
        elif filter_name in ("availabilityZone", "availability-zone"):
            return self.availability_zone
        elif filter_name in ("defaultForAz", "default-for-az"):
            return self.default_for_az
        elif filter_name == "state":
            return self.state
        else:
            return super().get_filter_value(filter_name, "DescribeSubnets")

    @classmethod
    def has_cfn_attr(cls, attr: str) -> bool:
        return attr in ["AvailabilityZone"]

    def get_cfn_attribute(self, attribute_name: str) -> None:
        from moto.cloudformation.exceptions import UnformattedGetAttTemplateException

        if attribute_name == "AvailabilityZone":
            raise NotImplementedError('"Fn::GetAtt" : [ "{0}" , "AvailabilityZone" ]"')
        raise UnformattedGetAttTemplateException()

    def get_available_subnet_ip(self, instance: Instance) -> str:
        try:
            new_ip = str(self._unused_ips.pop())
        except KeyError:
            new_ip_v4 = next(iter(self._subnet_ip_generator))

            # Skips any IP's if they've been manually specified
            while str(new_ip_v4) in self._subnet_ips:
                new_ip_v4 = next(iter(self._subnet_ip_generator))

            if new_ip_v4 == self.cidr.broadcast_address:
                raise StopIteration()  # Broadcast address cant be used obviously
            new_ip = str(new_ip_v4)
        # TODO StopIteration will be raised if no ip's available, not sure how aws handles this.

        self._subnet_ips[new_ip] = instance

        return new_ip

    # EFS calls this method as request_ip(str, MountTarget)
    # So technically it's not just Instances that are stored
    def request_ip(self, ip: str, instance: Instance) -> None:
        if ipaddress.ip_address(ip) not in self.cidr:
            raise Exception(f"IP does not fall in the subnet CIDR of {self.cidr}")

        if ip in self._subnet_ips:
            raise Exception("IP already in use")

        try:
            self._unused_ips.remove(ip)
        except KeyError:
            pass

        self._subnet_ips[ip] = instance

    def del_subnet_ip(self, ip: str) -> None:
        try:
            del self._subnet_ips[ip]
            self._unused_ips.add(ip)
        except KeyError:
            pass  # Unknown IP

    def attach_ipv6_cidr_block_associations(
        self, ipv6_cidr_block: str
    ) -> dict[str, Any]:
        association = {
            "associationId": random_subnet_ipv6_cidr_block_association_id(),
            "ipv6CidrBlock": ipv6_cidr_block,
            "ipv6CidrBlockState": {"State": "associated"},
        }
        self.ipv6_cidr_block_associations[str(association["associationId"])] = (
            association
        )
        return association

    def detach_subnet_cidr_block(self, association_id: str) -> dict[str, Any]:
        association = self.ipv6_cidr_block_associations.get(association_id)
        assert association is not None
        association["ipv6CidrBlockState"] = {"State": "disassociated"}
        return association

    def create_cidr_reservation(
        self,
        reservation_type: Literal["explicit"] | Literal["prefix"],
        cidr_block: ipaddress.IPv4Network | ipaddress.IPv6Network,
        tag_specifications: Optional[list[dict[str, Any]]] = None,
    ) -> SubnetCidrReservation:
        if not self._cidr_within_subnet_cidr_blocks(cidr_block):
            raise InvalidCidrReservationNotWithinSubnetCidr(str(self.cidr))

        if self._cidr_overlaps_with_existing_reservation(cidr_block):
            raise InvalidCidrReservationOverlapExisting(str(cidr_block))

        reservation = SubnetCidrReservation(
            ec2_backend=self.ec2_backend,
            subnet_id=self.id,
            cidr=cidr_block,
            reservation_type=reservation_type,
            owner_id=self.owner_id,
            tag_specifications=tag_specifications,
            subnet_cidr_reservation_id=random_subnet_cidr_reservation_id(),
        )

        for ip in cidr_block:
            self._reserved_ips.add(ip)

        if reservation_type == "prefix":
            # prefix reservations impact the avail ip count immediately
            self._available_ip_addresses -= cidr_block.num_addresses

        self._cidr_reservations.append(reservation)

        return reservation

    @cache
    def _base_avail_ip_iterator(
        self,
    ) -> (
        Iterator[ipaddress.IPv4Address]
        | list[ipaddress.IPv4Address]
        | Iterator[ipaddress.IPv6Address]
        | list[ipaddress.IPv6Address]
    ):
        return self.cidr.hosts()

    @property
    def _subnet_ip_generator(
        self,
    ) -> Iterator[ipaddress.IPv4Address | ipaddress.IPv6Address]:
        for host in self._base_avail_ip_iterator():
            # Skip any reserved ips
            if host in self._reserved_ips:
                continue
            yield host

    def _cidr_within_subnet_cidr_blocks(
        self, cidr_to_reserve: ipaddress.IPv4Network | ipaddress.IPv6Network
    ) -> bool:
        if isinstance(cidr_to_reserve, ipaddress.IPv4Network):
            if isinstance(self.cidr, ipaddress.IPv4Network):
                return cidr_to_reserve.subnet_of(self.cidr)
            return False

        for _, ipv6_block_association in self.ipv6_cidr_block_associations.items():
            ipv6_block = ipaddress.ip_network(ipv6_block_association["ipv6CidrBlock"])
            if isinstance(
                ipv6_block, ipaddress.IPv6Network
            ) and cidr_to_reserve.subnet_of(ipv6_block):
                return True
        return False

    def _cidr_overlaps_with_existing_reservation(
        self, cidr_to_reserve: ipaddress.IPv4Network | ipaddress.IPv6Network
    ) -> bool:
        cidr_to_reserve_is_ipv4: bool = isinstance(
            cidr_to_reserve, ipaddress.IPv4Network
        )
        for reservation in self._cidr_reservations:
            existing_block = reservation.cidr
            if (
                isinstance(existing_block, ipaddress.IPv4Network)
                and cidr_to_reserve_is_ipv4
                and cidr_to_reserve.overlaps(existing_block)
            ):
                return True
            elif (
                isinstance(existing_block, ipaddress.IPv6Network)
                and not cidr_to_reserve_is_ipv4
                and cidr_to_reserve.overlaps(existing_block)
            ):
                return True

        return False

    def delete_subnet_cidr_reservation(
        self,
        reservation_id: str,
    ) -> Optional[SubnetCidrReservation]:
        for index, reservation in enumerate(self._cidr_reservations):
            if reservation.id == reservation_id:
                self._cidr_reservations.pop(index)

                if reservation.reservation_type == "prefix":
                    for ip in reservation.cidr:
                        if (
                            ip not in self._subnet_ips
                            and ip not in self.aws_reserved_ips
                        ):
                            self._available_ip_addresses += 1
                return reservation

        return None


class SubnetRouteTableAssociation(CloudFormationModel):
    def __init__(self, route_table_id: str, subnet_id: str):
        self.route_table_id = route_table_id
        self.subnet_id = subnet_id

    @property
    def physical_resource_id(self) -> str:
        return self.route_table_id

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnetroutetableassociation.html
        return "AWS::EC2::SubnetRouteTableAssociation"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> SubnetRouteTableAssociation:
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        route_table_id = properties["RouteTableId"]
        subnet_id = properties["SubnetId"]

        ec2_backend = ec2_backends[account_id][region_name]
        subnet_association = ec2_backend.create_subnet_association(
            route_table_id=route_table_id, subnet_id=subnet_id
        )
        return subnet_association


class SubnetBackend:
    def __init__(self) -> None:
        # maps availability zone to dict of (subnet_id, subnet)
        self.subnets: dict[str, dict[str, Subnet]] = defaultdict(dict)
        self.subnet_associations: dict[str, SubnetRouteTableAssociation] = {}

    def get_subnet(self, subnet_id: str) -> Subnet:
        for subnets_per_zone in self.subnets.values():
            if subnet_id in subnets_per_zone:
                return subnets_per_zone[subnet_id]
        raise InvalidSubnetIdError(subnet_id)

    def get_default_subnets(self) -> dict[str, Subnet]:
        return {
            subnet.availability_zone: subnet
            for subnet in self.describe_subnets()
            if subnet.default_for_az
        }

    def create_default_subnet(self, availability_zone: str) -> Subnet:
        default_subnets = self.get_default_subnets()
        if availability_zone in default_subnets:
            raise DefaultSubnetAlreadyExistsInAvailabilityZoneError(
                default_subnets[availability_zone].id, availability_zone
            )

        default_vpc = self.get_default_vpc()  # type: ignore[attr-defined]
        if default_vpc is None:
            raise DefaultVpcDoesNotExistError()

        cidr_block = default_vpc.cidr_block
        blocks = list(ipaddress.ip_network(cidr_block).subnets(new_prefix=20))
        subnet = self.create_subnet(
            vpc_id=default_vpc.id,
            cidr_block=str(blocks[len(default_subnets)]),
            availability_zone=availability_zone,
        )
        return subnet

    def create_subnet(
        self,
        vpc_id: str,
        cidr_block: str,
        ipv6_cidr_block: Optional[str] = None,
        availability_zone: Optional[str] = None,
        availability_zone_id: Optional[str] = None,
        ipv6_native: bool = False,
        tags: Optional[dict[str, dict[str, str]]] = None,
    ) -> Subnet:
        subnet_id = random_subnet_id()
        # Validate VPC exists and the supplied CIDR block is a subnet of the VPC's
        vpc = self.get_vpc(vpc_id)  # type: ignore[attr-defined]
        if cidr_block:
            vpc_cidr_blocks = [
                ipaddress.IPv4Network(
                    str(cidr_block_association["cidr_block"]), strict=False
                )
                for cidr_block_association in vpc.get_cidr_block_association_set()
            ]
            try:
                subnet_cidr_block = ipaddress.IPv4Network(str(cidr_block), strict=False)
            except ValueError:
                raise InvalidCIDRBlockParameterError(cidr_block)

            subnet_in_vpc_cidr_range = False
            for vpc_cidr_block in vpc_cidr_blocks:
                if (
                    vpc_cidr_block.network_address <= subnet_cidr_block.network_address
                    and vpc_cidr_block.broadcast_address
                    >= subnet_cidr_block.broadcast_address
                ):
                    subnet_in_vpc_cidr_range = True
                    break

            if not subnet_in_vpc_cidr_range:
                raise InvalidSubnetRangeError(cidr_block)

            for subnet in self.describe_subnets(filters={"vpc-id": vpc_id}):
                if subnet.cidr.overlaps(subnet_cidr_block):
                    raise InvalidSubnetConflictError(cidr_block)

        # if this is the first subnet for an availability zone,
        # consider it the default
        default_for_az = len(self.subnets.get(availability_zone, [])) == 0  # type: ignore[arg-type]
        map_public_ip_on_launch = default_for_az

        if availability_zone is None and not availability_zone_id:
            availability_zone = "us-east-1a"
        try:
            if availability_zone:
                availability_zone_data = next(
                    zone
                    for zones in RegionsAndZonesBackend.zones.values()
                    for zone in zones
                    if zone.name == availability_zone
                )
            elif availability_zone_id:
                availability_zone_data = next(
                    zone
                    for zones in RegionsAndZonesBackend.zones.values()
                    for zone in zones
                    if zone.zone_id == availability_zone_id
                )

        except StopIteration:
            raise InvalidAvailabilityZoneError(
                availability_zone,
                ", ".join(
                    [
                        zone.name
                        for zones in RegionsAndZonesBackend.zones.values()
                        for zone in zones
                    ]
                ),
            )
        subnet = Subnet(
            self,
            subnet_id,
            vpc_id,
            cidr_block,
            ipv6_cidr_block,
            availability_zone_data,
            default_for_az,
            map_public_ip_on_launch,
            ipv6_native=ipv6_native,
        )

        for k, v in tags.get("subnet", {}).items() if tags else []:
            subnet.add_tag(k, v)

        # AWS associates a new subnet with the default Network ACL
        self.associate_default_network_acl_with_subnet(subnet_id, vpc_id)  # type: ignore[attr-defined]
        self.subnets[availability_zone][subnet_id] = subnet  # type: ignore[index]
        return subnet

    def describe_subnets(
        self, subnet_ids: Optional[list[str]] = None, filters: Optional[Any] = None
    ) -> list[Subnet]:
        # Extract a list of all subnets
        matches = list(
            itertools.chain(*[x.copy().values() for x in self.subnets.copy().values()])
        )
        if subnet_ids:
            matches = [sn for sn in matches if sn.id in subnet_ids]
            if len(subnet_ids) > len(matches):
                unknown_ids = set(subnet_ids) - set(matches)  # type: ignore[arg-type]
                raise InvalidSubnetIdError(list(unknown_ids)[0])
        if filters:
            matches = generic_filter(filters, matches)

        return matches

    def delete_subnet(self, subnet_id: str) -> Subnet:
        for subnets in self.subnets.values():
            if subnet_id in subnets:
                return subnets.pop(subnet_id)
        raise InvalidSubnetIdError(subnet_id)

    def modify_subnet_attribute(
        self, subnet_id: str, attr_name: str, attr_value: str
    ) -> None:
        subnet = self.get_subnet(subnet_id)
        if attr_name in ("map_public_ip_on_launch", "assign_ipv6_address_on_creation"):
            setattr(subnet, attr_name, attr_value)
        else:
            raise InvalidParameterValueError(attr_name)

    def get_subnet_from_ipv6_association(self, association_id: str) -> Optional[Subnet]:
        subnet = None
        for s in self.describe_subnets():
            if association_id in s.ipv6_cidr_block_associations:
                subnet = s
        return subnet

    def associate_subnet_cidr_block(
        self, subnet_id: str, ipv6_cidr_block: str
    ) -> dict[str, Any]:
        subnet = self.get_subnet(subnet_id)
        if not subnet:
            raise InvalidSubnetIdError(subnet_id)
        association = subnet.attach_ipv6_cidr_block_associations(ipv6_cidr_block)
        return association

    def disassociate_subnet_cidr_block(
        self, association_id: str
    ) -> tuple[str, dict[str, str]]:
        subnet = self.get_subnet_from_ipv6_association(association_id)
        if not subnet:
            raise InvalidSubnetCidrBlockAssociationID(association_id)
        association = subnet.detach_subnet_cidr_block(association_id)
        return subnet.id, association

    def create_subnet_association(
        self, route_table_id: str, subnet_id: str
    ) -> SubnetRouteTableAssociation:
        subnet_association = SubnetRouteTableAssociation(route_table_id, subnet_id)
        self.subnet_associations[f"{route_table_id}:{subnet_id}"] = subnet_association
        return subnet_association

    def create_subnet_cidr_reservation(
        self,
        subnet_id: str,
        reservation_type: str,
        cidr: str,
        tags: Optional[list[dict[str, Any]]] = None,
    ) -> SubnetCidrReservation:
        if reservation_type not in {"prefix", "explicit"}:
            raise InvalidPrefixReservationType(reservation_type)

        reservation_type_lit: Literal["prefix", "explicit"] = cast(
            Literal["prefix", "explicit"], reservation_type
        )

        cidr_to_reserve: ipaddress.IPv4Network | ipaddress.IPv6Network
        try:
            cidr_to_reserve = ipaddress.ip_network(cidr)
        except ValueError:
            raise InvalidCIDRBlockParameterError(cidr)

        subnet = self.get_subnet(subnet_id)

        scr = subnet.create_cidr_reservation(
            reservation_type_lit, cidr_to_reserve, tags
        )
        return scr

    def get_subnet_cidr_reservations(
        self,
        subnet_id: str,
        filters: Optional[Any] = None,
    ) -> dict[str, list[SubnetCidrReservation]]:
        matches: list[SubnetCidrReservation] = self.get_subnet(
            subnet_id
        )._cidr_reservations

        if filters:
            matches = generic_filter(filters, matches)

        ipv4_reservations = []
        ipv6_reservations = []

        for reservation in matches:
            if isinstance(reservation.cidr, ipaddress.IPv4Network):
                ipv4_reservations.append(reservation)
            else:
                ipv6_reservations.append(reservation)

        return {
            "SubnetIpv4CidrReservations": ipv4_reservations,
            "SubnetIpv6CidrReservations": ipv6_reservations,
        }

    def delete_subnet_cidr_reservation(
        self,
        reservation_id: str,
    ) -> SubnetCidrReservation:
        for az_subnet in self.subnets.values():
            for subnet in az_subnet.values():
                deleted_reservation = subnet.delete_subnet_cidr_reservation(
                    reservation_id
                )
                if deleted_reservation:
                    return deleted_reservation

        raise InvalidCidrReservationNotFound(reservation_id)



import re
from typing import Any, Final, Optional

from jsonpath_ng.ext import parse
from jsonpath_ng.jsonpath import Index

from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

_PATTERN_SINGLETON_ARRAY_ACCESS_OUTPUT: Final[str] = r"\[\d+\]$"
_PATTERN_SLICE_OR_WILDCARD_ACCESS = (
    r"\$(?:\.[^[]+\[(?:\*|\d*:\d*)\]|\[\*\])(?:\.[^[]+)*$"
)


def _is_singleton_array_access(path: str) -> bool:
    # Returns true if the json path terminates with a literal singleton array access.
    return bool(re.search(_PATTERN_SINGLETON_ARRAY_ACCESS_OUTPUT, path))


def _contains_slice_or_wildcard_array(path: str) -> bool:
    # Returns true if the json path contains a slice or wildcard in the array.
    # Slices at the root are discarded, but wildcard at the root is allowed.
    return bool(re.search(_PATTERN_SLICE_OR_WILDCARD_ACCESS, path))


class NoSuchJsonPathError(Exception):
    json_path: Final[str]
    data: Final[Any]
    _message: Optional[str]

    def __init__(self, json_path: str, data: Any):
        self.json_path = json_path
        self.data = data
        self._message = None

    @property
    def message(self) -> str:
        if self._message is None:
            data_json_str = to_json_str(self.data)
            self._message = f"The JSONPath '{self.json_path}' could not be found in the input '{data_json_str}'"
        return self._message

    def __str__(self):
        return self.message


def extract_json(path: str, data: Any) -> Any:
    input_expr = parse(path)

    matches = input_expr.find(data)
    if not matches:
        if _contains_slice_or_wildcard_array(path):
            return []
        raise NoSuchJsonPathError(json_path=path, data=data)

    if (
        len(matches) > 1
        or isinstance(matches[0].path, Index)
        # Last condition is different from LS and fixes a very specific bug
        # https://github.com/getmoto/moto/issues/7825
        or (matches[0].context and isinstance(matches[0].context.path, Index))
    ):
        value = [match.value for match in matches]

        # AWS StepFunctions breaks jsonpath specifications and instead
        # unpacks literal singleton array accesses.
        if _is_singleton_array_access(path=path) and len(value) == 1:
            value = value[0]
    else:
        value = matches[0].value

    return value



from collections import defaultdict
from datetime import datetime
from typing import Any, Optional

from moto.ec2.models.spot_requests import SpotFleetLaunchSpec, SpotInstanceRequest

from ...core.utils import utcnow
from ..utils import (
    convert_tag_spec,
    random_fleet_id,
)
from .core import TaggedEC2Resource


class Fleet(TaggedEC2Resource):
    def __init__(
        self,
        ec2_backend: Any,
        fleet_id: str,
        on_demand_options: dict[str, Any],
        spot_options: dict[str, Any],
        target_capacity_specification: dict[str, Any],
        launch_template_configs: list[dict[str, Any]],
        excess_capacity_termination_policy: str,
        replace_unhealthy_instances: bool,
        terminate_instances_with_expiration: bool,
        fleet_type: str,
        valid_from: Optional[datetime],
        valid_until: Optional[datetime],
        tag_specifications: list[dict[str, Any]],
    ):
        self.ec2_backend = ec2_backend
        self.id = fleet_id
        self.spot_options = spot_options
        self.on_demand_options = on_demand_options
        self.target_capacity_specification = target_capacity_specification
        self.launch_template_configs = launch_template_configs
        self.excess_capacity_termination_policy = (
            excess_capacity_termination_policy or "termination"
        )
        self.replace_unhealthy_instances = replace_unhealthy_instances
        self.terminate_instances_with_expiration = terminate_instances_with_expiration
        self.fleet_type = fleet_type
        self.valid_from = valid_from or utcnow()
        self.valid_until = valid_until
        tag_spec = convert_tag_spec(tag_specifications)
        self.add_tags(tag_spec.get("fleet", {}))
        self.tags = self.get_tags()
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/create_fleet.html
        # If the fleet type is instant, specify a resource type of fleet to tag the fleet or instance to tag the instances at launch.
        # If the fleet type is maintain or request, specify a resource type of fleet to tag the fleet. You cannot specify a resource type of instance. To tag instances at launch, specify the tags in a launch template.
        instance_tags = (
            tag_spec.get("instance", {}) if self.fleet_type == "instant" else {}
        )

        self.state = "active"
        self.fulfilled_capacity = 0.0
        self.fulfilled_on_demand_capacity = 0.0
        self.fulfilled_spot_capacity = 0.0

        self.launch_specs: list[SpotFleetLaunchSpec] = []

        for config in launch_template_configs or []:
            launch_spec = config["LaunchTemplateSpecification"]
            if "LaunchTemplateId" in launch_spec:
                launch_template = self.ec2_backend.get_launch_template(
                    template_id=launch_spec["LaunchTemplateId"]
                )
            elif "LaunchTemplateName" in launch_spec:
                launch_template = self.ec2_backend.get_launch_template_by_name(
                    name=launch_spec["LaunchTemplateName"]
                )
            else:
                continue

            # Resolve $Latest or $Default to actual version number
            resolved_launch_spec = launch_spec.copy()
            if resolved_launch_spec.get("Version") == "$Latest":
                resolved_launch_spec["Version"] = str(
                    launch_template.latest_version_number
                )
            elif resolved_launch_spec.get("Version") == "$Default":
                resolved_launch_spec["Version"] = str(
                    launch_template.default_version_number
                )
            # Always include the template ID in response (AWS does this even when name is used)
            resolved_launch_spec["LaunchTemplateId"] = launch_template.id

            launch_template_data = launch_template.latest_version().data
            overrides_list = config.get("Overrides") or [{}]
            for override in overrides_list:
                # Merge launch template data with override
                spec = launch_template_data.copy()
                spec.update(override)

                tag_spec_set = spec.get("TagSpecifications", [])
                tags = convert_tag_spec(tag_spec_set)
                tags["instance"] = tags.get("instance", {}) | instance_tags

                self.launch_specs.append(
                    SpotFleetLaunchSpec(
                        ebs_optimized=spec.get("EbsOptimized"),
                        group_set=spec.get("GroupSet", []),
                        iam_instance_profile=spec.get("IamInstanceProfile"),
                        image_id=spec["ImageId"],
                        instance_type=spec["InstanceType"],
                        key_name=spec.get("KeyName"),
                        monitoring=spec.get("Monitoring"),
                        spot_price=spec.get("SpotPrice"),
                        subnet_id=spec.get("SubnetId"),
                        tag_specifications=tags,
                        user_data=spec.get("UserData"),
                        weighted_capacity=spec.get("WeightedCapacity", 1),
                        launch_template_spec=resolved_launch_spec,
                        overrides=override if override else None,
                    )
                )

        self.spot_requests: list[SpotInstanceRequest] = []
        self.on_demand_instances: list[dict[str, Any]] = []
        default_capacity = (
            target_capacity_specification.get("DefaultTargetCapacityType")
            or "on-demand"
        )
        self.target_capacity = int(
            target_capacity_specification.get("TotalTargetCapacity")  # type: ignore[arg-type]
        )
        self.spot_target_capacity = int(
            target_capacity_specification.get("SpotTargetCapacity", 0)
        )
        if self.spot_target_capacity > 0:
            self.create_spot_requests(self.spot_target_capacity)
        self.on_demand_target_capacity = int(
            target_capacity_specification.get("OnDemandTargetCapacity", 0)
        )
        if self.on_demand_target_capacity > 0:
            self.create_on_demand_requests(self.on_demand_target_capacity)

        remaining_capacity = self.target_capacity - self.fulfilled_capacity
        if remaining_capacity > 0:
            if default_capacity == "on-demand":
                self.create_on_demand_requests(remaining_capacity)
            elif default_capacity == "spot":
                self.create_spot_requests(remaining_capacity)

    @property
    def type(self) -> str:
        return self.fleet_type

    @property
    def valid_from_as_string(self) -> str:
        x = self.valid_from
        return f"{x.year}-{x.month:02d}-{x.day:02d}T{x.hour:02d}:{x.minute:02d}:{x.second:02d}.000Z"

    @property
    def valid_until_as_string(self) -> Optional[str]:
        if self.valid_until is None:
            return self.valid_until
        x = self.valid_until
        return f"{x.year}-{x.month:02d}-{x.day:02d}T{x.hour:02d}:{x.minute:02d}:{x.second:02d}.000Z"

    @property
    def physical_resource_id(self) -> str:
        return self.id

    @property
    def instances(self) -> Optional[list[dict[str, Any]]]:
        """
        Return instances for instant fleets, None for other fleet types.
        This is part of the CreateFleet response for instant fleets only.
        """
        if self.fleet_type != "instant":
            return None

        instances = []

        # Process on-demand instances
        for item in self.on_demand_instances:
            instance_data = self._build_instance_data(
                instance=item["instance"],
                lifecycle="on-demand",
                launch_spec=item.get("launch_spec"),
            )
            instances.append(instance_data)

        # Process spot instances
        for spot_request in self.spot_requests:
            instance_data = self._build_instance_data(
                instance=spot_request.instance,
                lifecycle="spot",
                launch_spec=spot_request.launch_spec,
            )
            instances.append(instance_data)

        return instances

    @staticmethod
    def _build_instance_data(
        instance: Any, lifecycle: str, launch_spec: Optional[SpotFleetLaunchSpec]
    ) -> dict[str, Any]:
        instance_data = {
            "Lifecycle": lifecycle,
            "InstanceIds": [instance.id],
            "InstanceType": instance.instance_type,
        }

        # Add launch template and overrides if available
        if launch_spec and launch_spec.launch_template_spec:
            launch_template_and_overrides: dict[str, Any] = {
                "LaunchTemplateSpecification": launch_spec.launch_template_spec
            }
            if launch_spec.overrides:
                launch_template_and_overrides["Overrides"] = launch_spec.overrides

            instance_data["LaunchTemplateAndOverrides"] = launch_template_and_overrides

        return instance_data

    def create_spot_requests(self, weight_to_add: float) -> list[SpotInstanceRequest]:
        weight_map, added_weight = self.get_launch_spec_counts(weight_to_add)
        for launch_spec, count in weight_map.items():
            requests = self.ec2_backend.request_spot_instances(
                price=launch_spec.spot_price,
                image_id=launch_spec.image_id,
                count=count,
                spot_instance_type="persistent",
                valid_from=None,
                valid_until=None,
                launch_group=None,
                availability_zone_group=None,
                key_name=launch_spec.key_name,
                security_groups=launch_spec.group_set,
                user_data=launch_spec.user_data,
                instance_type=launch_spec.instance_type,
                placement=None,
                kernel_id=None,
                ramdisk_id=None,
                monitoring_enabled=launch_spec.monitoring,
                subnet_id=launch_spec.subnet_id,
                spot_fleet_id=self.id,
                tags=launch_spec.tag_specifications,
                launch_spec=launch_spec,
            )
            self.spot_requests.extend(requests)
        self.fulfilled_capacity += added_weight
        return self.spot_requests

    def create_on_demand_requests(self, weight_to_add: float) -> None:
        weight_map, added_weight = self.get_launch_spec_counts(weight_to_add)
        for launch_spec, count in weight_map.items():
            reservation = self.ec2_backend.run_instances(
                image_id=launch_spec.image_id,
                count=count,
                instance_type=launch_spec.instance_type,
                is_instance_type_default=False,
                key_name=launch_spec.key_name,
                security_group_names=launch_spec.group_set,
                user_data=launch_spec.user_data,
                placement=None,
                kernel_id=None,
                ramdisk_id=None,
                monitoring_enabled=launch_spec.monitoring,
                subnet_id=launch_spec.subnet_id,
                fleet_id=self.id,
                tags=launch_spec.tag_specifications,
            )

            for instance in reservation.instances:
                self.on_demand_instances.append(
                    {
                        "id": reservation.id,
                        "instance": instance,
                        "launch_spec": launch_spec,
                    }
                )
        self.fulfilled_capacity += added_weight

    def get_launch_spec_counts(
        self, weight_to_add: float
    ) -> tuple[dict[SpotFleetLaunchSpec, int], float]:
        weight_map: dict[SpotFleetLaunchSpec, int] = defaultdict(int)

        weight_so_far = 0.0
        if (
            self.spot_options
            and self.spot_options["AllocationStrategy"] == "diversified"
        ):
            launch_spec_index = 0
            while True:
                launch_spec = self.launch_specs[
                    launch_spec_index % len(self.launch_specs)
                ]
                weight_map[launch_spec] += 1
                weight_so_far += launch_spec.weighted_capacity
                if weight_so_far >= weight_to_add:
                    break
                launch_spec_index += 1
        else:  # lowestPrice
            cheapest_spec = sorted(
                # FIXME: change `+inf` to the on demand price scaled to weighted capacity when it's not present
                self.launch_specs,
                key=lambda spec: float(spec.spot_price or "+inf"),
            )[0]
            weight_so_far = weight_to_add + (
                weight_to_add % cheapest_spec.weighted_capacity
            )
            weight_map[cheapest_spec] = int(
                weight_so_far // cheapest_spec.weighted_capacity
            )

        return weight_map, weight_so_far

    def terminate_instances(self) -> None:
        instance_ids = []
        new_fulfilled_capacity = self.fulfilled_capacity
        for req in self.spot_requests + self.on_demand_instances:
            instance = None
            try:
                instance = req.instance  # type: ignore
            except AttributeError:
                instance = req["instance"]  # type: ignore[index]

            if instance.state == "terminated":
                continue

            # stop when we hit the target capacity
            if new_fulfilled_capacity <= self.target_capacity:
                break

            instance_ids.append(instance.id)
            new_fulfilled_capacity -= 1

        self.spot_requests = [
            req for req in self.spot_requests if req.instance.id not in instance_ids
        ]
        self.on_demand_instances = [
            req
            for req in self.on_demand_instances
            if req["instance"].id not in instance_ids
        ]
        self.ec2_backend.terminate_instances(instance_ids)


class FleetsBackend:
    def __init__(self) -> None:
        self.fleets: dict[str, Fleet] = {}

    def create_fleet(
        self,
        on_demand_options: dict[str, Any],
        spot_options: dict[str, Any],
        target_capacity_specification: dict[str, Any],
        launch_template_configs: list[dict[str, Any]],
        excess_capacity_termination_policy: str,
        replace_unhealthy_instances: bool,
        terminate_instances_with_expiration: bool,
        fleet_type: str,
        valid_from: Optional[datetime],
        valid_until: Optional[datetime],
        tag_specifications: list[dict[str, Any]],
    ) -> Fleet:
        fleet_id = random_fleet_id()
        fleet = Fleet(
            self,
            fleet_id,
            on_demand_options,
            spot_options,
            target_capacity_specification,
            launch_template_configs,
            excess_capacity_termination_policy,
            replace_unhealthy_instances,
            terminate_instances_with_expiration,
            fleet_type,
            valid_from,
            valid_until,
            tag_specifications,
        )
        self.fleets[fleet_id] = fleet
        return fleet

    def get_fleet(self, fleet_id: str) -> Optional[Fleet]:
        return self.fleets.get(fleet_id)

    def describe_fleet_instances(self, fleet_id: str) -> list[Any]:
        fleet = self.get_fleet(fleet_id)
        if not fleet:
            return []
        # TODO: These are incompatible types (list[object] + list[dict]) and should be normalized.
        return fleet.spot_requests + fleet.on_demand_instances

    def describe_fleets(self, fleet_ids: Optional[list[str]]) -> list[Fleet]:
        fleets = list(self.fleets.values())

        if fleet_ids:
            fleets = [fleet for fleet in fleets if fleet.id in fleet_ids]

        return fleets

    def delete_fleets(
        self, fleet_ids: list[str], terminate_instances: bool
    ) -> list[Fleet]:
        fleets = []
        for fleet_id in fleet_ids:
            fleet = self.fleets[fleet_id]
            if terminate_instances:
                # State indicates the fleet is in the process of being terminated
                # AWS will change the state to `deleted` after a few seconds/minutes
                # Note that it will stay in the `deleted`-state for at least a few hours
                fleet.state = "deleted_terminating"
                fleet.target_capacity = 0
                fleet.terminate_instances()
            else:
                # State is different, and indicates that instances are still running
                fleet.state = "deleted_running"
            fleets.append(fleet)
        return fleets



from typing import Any

from moto.core.request import normalize_request
from moto.elb.responses import ELBResponse
from moto.elbv2.responses import ELBV2Response


def api_version_elb_backend(*args: Any, **kwargs: Any) -> Any:
    """
    ELB and ELBV2 (Classic and Application load balancers) use the same
    hostname and url space. To differentiate them we must read the
    `Version` parameter out of the url-encoded request body. TODO: There
    has _got_ to be a better way to do this. Please help us think of
    one.
    """
    request = normalize_request(args[0])
    version = request.values.get("Version")
    if "2012-06-01" == version:
        return ELBResponse.dispatch(*args, **kwargs)
    elif "2015-12-01" == version:
        return ELBV2Response.dispatch(*args, **kwargs)
    else:
        raise Exception(f"Unknown ELB API version: {version}")


url_bases = [r"https?://elasticloadbalancing\.(.+)\.amazonaws.com"]

url_paths = {"{0}/$": api_version_elb_backend}
